/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Abstraction layer for Gmail response data. glDataObject parses input data and
 * provides the extracted data back through a common interface. Input can be the
 * VIEW_DATA variable (when set) or request responses (js or html).<br/>
 * <br/>
 * When <code>glDataObject.createFromData(data)</code> is called, the usable raw data is stored
 * in the _rawData property of the created glDataObject object. Then, when the data
 * is requested, the extraction is run and the results cached. This way we increase
 * performance and reduce unnecessary memory churn (or so I hope).
 *
 * @class ?
 */
var glDataObject = glBase.extend({
    // Properties {
    /**
     * The un-processed raw data
     * @property {Object} ?
     */
    _rawData: null,

    /**
     * The extracted user label data
     * @property {Array} ?
     */
    _userLabels: null,

    /**
     * The extracted internal label data
     * @property {Array} ?
     */
    _internalLabels: null,

    /**
     * The extracted additional accounts
     * @property {Array} ?
     */
    _additionalAccounts: null,

    /**
     * The extracted query data
     * @property {Object} ?
     */
    _queryData: null,

    /**
     * The extracted threads
     * @property {glThread[]} ?
     */
    _threads: null,

	/** The extracted thread data
	* @property {Object} ?
	*/
	_threadData: null,

	/** The extracted thread messages
	* @property {glThreadMessage[]}
	*/
	_threadMessages: null,

    /**
     * The gmail object for the page
     * @property {private glGmail} ?
     */
    _gmail: null,
    //}

    /**
     * The constructor.
     * @constructor {private} ?
	 * @param {glGmail} gmail
     */
    constructor: function(gmail) {
        this._rawData = {};
		this._gmail = gmail;
    },

    // Methods {
    /**
     * Returns the extracted additional account data
     * @function {Array} ?
     * @return A list of additional e-mail addresses the user can use and their display names
     */
    getAdditionalAccounts: function() {
        if(this._additionalAccounts == null) {
		 	// Data not yet extracted, so extract and save it
            this._additionalAccounts = [];
			var rawData = this._rawData[glDataObject.ADDITIONAL_ACCOUNTS_TAG];
			if(rawData) {
                for(var j = 0; j < rawData[1].length; j++) {
                    this._additionalAccounts.push({
                        name: rawData[1][j][0],
                        email: rawData[1][j][1]
                    });
                }
            }

            // Clean up references to raw data, since it's not needed anymore
            delete this._rawData[glDataObject.ADDITIONAL_ACCOUNTS_TAG];
            rawData = null;
        }
        return this._additionalAccounts;
    },

    /**
     * Returns the extracted list of the user-created labels
     * @function {Array} ?
     * @return A list of the user-created labels and data about them (like threads, unread threads, etc.)
     */
    getUserLabels: function() {
        if(this._userLabels == null) {
            // Data not yet extracted, so extract and save it
            this._userLabels = [];
            var rawData = this._rawData[glDataObject.LABEL_DATA_TAG];
            if(rawData) {
                for(var j = 0; j < rawData[2].length; j++) {
                    this._userLabels.push({
                        name: rawData[2][j][0],
                        unreadThreads: rawData[2][j][1],
                        threads: rawData[2][j][2],
                    });
                }

                // Clean up references to raw data, since it's not needed anymore
                if(this._internalLabels !== null) {
                    // Both internal and user labels extracted so delete all label data
                    delete this._rawData[glDataObject.LABEL_DATA_TAG];
                }
            }

            rawData = null;
        }
        return this._userLabels;
    },

    /**
     * Returns the extracted list of the internal labels (inbox, starred, spam, etc.)
     * @function {Array} ?
     * @return A list of the internal labels and the number of unread threads in them
     */
    getInternalLabels: function() {
        if(this._internalLabels == null) {
            // Data not yet extracted, so extract and save it
            this._internalLabels = [];
            var rawData = this._rawData[glDataObject.LABEL_DATA_TAG];
            if(rawData) {
                for(var j = 0; j < rawData[1].length; j++) {
                    this._internalLabels.push({
                        name: rawData[1][j][0],
                        unreadThreads: rawData[1][j][1],
                    });
                }

                // Clean up references to raw data, since it's not needed anymore
                if(this._userLabels !== null) {
                    // Both internal and user labels extracted so delete all label data
                    delete this._rawData[glDataObject.LABEL_DATA_TAG];
                }
            }

            rawData = null;
        }
        return this._internalLabels;
    },

    /**
     * Returns the extracted action response data, which typically manifests in
     * the UI as the yellow box at the top. This chunk contains the affected thread
     * id, which is used to determine the thread id of new messages.
     * @function {Object} ?
     * @return The action response data, which includes the message and the id of the affected thread if available
     */
    getActionResponseData: function() {
        if(this._actionResponseData == null) {
            this._actionResponseData = {};
            var rawData = this._rawData[glDataObject.ACTIVITY_RESPONSE_TAG];
            if(rawData) {
                this._actionResponseData.message = rawData[2];
                this._actionResponseData.affectedThread = (rawData.length > 3) ? rawData[3][0] : null;
            }

            // Clean up references to raw data, since it's not needed anymore
            delete this._rawData[glDataObject.ACTIVITY_RESPONSE_TAG];
            rawData = null;
        }
        return this._actionResponseData;
    },

    /**
     * Returns the extracted query data (number of results and query string)
     * @function {Object} ?
     * @return A hash containing the query string and the number of results
     */
    getQueryData: function() {
        if(this._queryData == null) {
            this._queryData = {};
            var rawData = this._rawData[glDataObject.QUERY_DATA_TAG];
            if(rawData) {
                this._queryData.query   = rawData[5];
                this._queryData.results = rawData[2];
            }

            // Clean up references to raw data, since it's not needed anymore
            delete this._rawData[glDataObject.QUERY_DATA_TAG];
            rawData = null;
        }
        return this._queryData;
    },

    /**
     * Returns the extracted threads
     * @function {glThread[]} ?
     * @return An array of the extracted threads wrapped in glThread objects
     */
    getThreads: function() {
        if(this._threads == null) {
            this._threads = [];
            var rawData = this._rawData[glDataObject.THREAD_DATA_TAG];
            if(rawData) {
                for each(var chunk in rawData) {
                    // Thread data starts at i = 3 (0 is the chunk name, 1 is the thread offset, and 2 is the thread count for the block)
                    for(var i = 3; i < chunk.length; i++) {
                    	var knownContacts = [];
                    	if( chunk[i][7] ){
		            	var m = chunk[i][7].match(/<span[^>]+?email=[^>]+?>[^<]+?<\/span>/gi);
				if( m ){
					for( var j = 0; j < m.length; j++ ){
						var m2 = m[j].match(/<span[^>]+?email=\"([^"]+?)\"[^>]*?>([^<]+?)<\/span>/i);
						knownContacts.push({name:m2[2], email:glBase.realEmail(m2[1])});
					}
				}
			}
                        this._threads.push(glThread.createFromData({
                            threadId: chunk[i][2],
                            isRead: (chunk[i][3] == 1),
                            isStarred: (chunk[i][4] == 1),
                            labels: chunk[i][5], // Array of labels - internal labels prefixed with a "^"
                            contactsSnippet: chunk[i][7], // Original e-mail line with HTML for view
                            knownContacts: knownContacts,
                            subjectSnippet: chunk[i][9],
                            bodySnippet: chunk[i][10],
                            attachmentsCSV: chunk[i][13], // Comma seperated but does not escape commas found in file names, so I'm not adjusting it
                            dateSnippet: chunk[i][14],
                            date: chunk[i][15],
                            timestamp: chunk[i][16],
                        }, this._gmail));
                    }
                }
            }

            // Clean up references to raw data, since it's not needed anymore
            delete this._rawData[glDataObject.THREAD_DATA_TAG];
            rawData = null;
        }
        return this._threads;
    },

	/**
	* Extract the thread data from a conversation ('ms') request.
	* @function {Object} ?
	* @param {optional Boolean} preserveData - If true, do not delete raw data
	* @return The thread data in a simple object
	*/
	getThreadData: function(preserveData) {
        if(this._threadData == null) {
            var rawData = this._rawData[glDataObject.THREAD_MESSAGE_DATA_TAG];
            if(rawData) {
            			var labelsHash = {};
				var emailsHash = {};
				var filesHash = {};
				for each(var chunk in rawData){
					if( chunk[13][9] && chunk[13][9][1] && chunk[13][9][1].length>0 ){
						for( var i = 0; i < chunk[13][9][1].length; i++ ) emailsHash[glBase.realEmail(chunk[13][9][1][i][1])] = {name:chunk[13][9][1][i][0], email:glBase.realEmail(chunk[13][9][1][i][1])};
					}else if( chunk[5] && chunk[6] ){
						emailsHash[glBase.realEmail(chunk[6])] = {name:chunk[5], email:glBase.realEmail(chunk[6])};
					}
					if( chunk[13][7] && chunk[13][7][0] ){
						for( var i = 0; i < chunk[13][7][0].length; i++ ) filesHash[chunk[13][7][0][i][1]] = true;
					}
					for( var i = 0; i < chunk[9].length; i++ ) labelsHash[chunk[9][i]] = true;	// It seems labels don't appear on all ms. e.g. apply S/Action half way through thread, and all subsequent messages don't carry it.0
				}

				var knownContacts = [];
				for( e in emailsHash ) knownContacts.push( emailsHash[e] );
				var chunk = rawData[rawData.length-1];	// Get the last chunk - most accurate re: date, etc.
				
				
				this._threadData = {
					labels: glBase.hashKeysToArray(labelsHash),
					subjectSnippet: glBase.decodeEntities(chunk[12].replace(/<.+?>/g, "")).replace(/\s{2,}/g, " "),
					dateSnippet: chunk[17],
					date: chunk[24],
					timestamp: chunk[7],
					knownContacts: knownContacts,
					attachmentsCSV: glBase.hashKeysToArray(filesHash).join(","),
				};
            }

            // Clean up references to raw data, since it's not needed anymore
			if( !preserveData ){
            	delete this._rawData[glDataObject.THREAD_MESSAGE_DATA_TAG];
            	rawData = null;
			}
        }
        return this._threadData;
    },

    /**
     * Returns the extracted thread messages
     * @function {glThreadMessage[]} ?
     * @return An array of the extracted threads wrapped in glThreadMessage objects
     */
	getThreadMessages: function(){
		if( this._threadMessages==null ){
			this._threadMessages = [];
			var rawData = this._rawData[glDataObject.THREAD_MESSAGE_DATA_TAG];
			if( rawData ){
				for each(var chunk in rawData){
					var recipients = null;
					if( chunk[13][9] && chunk[13][9][1] && chunk[13][9][1].length>0 ){
						recipients = [];
						for( var i = 0; i < chunk[13][9][1].length; i++ ){
							recipients.push({
								name: chunk[13][9][1][i][0],
								email: glBase.realEmail(chunk[13][9][1][i][1]),
							})
						}
					}
					var sender = null;
					if( chunk[6] ){
						sender = {name:chunk[5],email:glBase.realEmail(chunk[6])};
					}
					var files = [];
					if( chunk[13][7] && chunk[13][7][0] ){
						for( var i = 0; i < chunk[13][7][0].length; i++ ){
							var f = chunk[13][7][0][i];
							files.push({
								attid: f[0],
								name: f[1],
								bytes: f[3],
								icon: f[6],
								url: f[9],
								type: f[7],
								preview: f[8],
							});
						}
					}
					
					this._threadMessages.push( glThreadMessage.createFromData({
						messageId: chunk[1],
						timestamp: chunk[7],
						message: chunk[13][6],
						date: chunk[13][10],
						sender: sender,
						recipients: recipients,
						files: files,
					}, this._gmail) );
				}
			}
			// Clean up references to raw data, since it's not needed anymore
           	delete this._rawData[glDataObject.THREAD_MESSAGE_DATA_TAG];
           	rawData = null;
		}
		return this._threadMessages;
	},
    //}
},{
    // Static Properties {
    ADDITIONAL_ACCOUNTS_TAG: 'cfs',
    LABEL_DATA_TAG: 'ld',
    LOGIN_ACTIVITY_TAG: 'la',
    QUERY_DATA_TAG: 'ti',
    THREAD_DATA_TAG: 'tb',
    THREAD_MESSAGE_DATA_TAG: 'ms',
    ACTIVITY_RESPONSE_TAG: 'a',
    //}

    // Static Methods {
    /**
     * Pass in the response text of a type "j" response, and it generates a
     * glDataObject object.
     * @function {static glDataObject} ?
     * @param {String} response - The reponse text to generate the glDataObject object from
     * @param {String} url - The URL for the request (used for the sandbox URL)
	 * @param {glGmail} gmail - 
     * @return The generated glDataObject object
     */
    createFromJSResponse: function(response, url, gmail) {
        if(typeof url == 'undefined') url = 'https://mail.google.com/';
        glLogger.log("Creating glDataObject object from JS response...", 'glDataObject');

        // Remove while(1) (which Gmail puts there for some reason)
        response = response.replace('while(1);', '');

        // Eval JS in sandbox and extract
        var evalJS = 'var result = (' + response + ')';
        var sb = Components.utils.Sandbox(url);
        Components.utils.evalInSandbox(evalJS, sb);
        if(typeof sb.result != 'object' ) throw new Error('result not an object');

        return glDataObject.createFromData(sb.result[0], gmail);
    },

    /**
     * Pass in the response text of a type "h" response, and it generates a
     * glDataObject object.
     * @function {static glDataObject} ?
     * @param {String} response - The reponse text to generate the glDataObject object from
     * @param {String} url - The URL for the request (used for the sandbox URL)
	 * @param {glGmail} gmail -
     * @return The generated glDataObject object
     */
    createFromHTMLResponse: function(response, url, gmail) {
        if(typeof url == 'undefined') url = 'https://mail.google.com/';
        glLogger.log("Creating glDataObject object from HTML response...", 'glDataObject');

        // Convert response HTML to proper XML for parsing
	// Nb The use of ,? is because Gmail seems to insert commas randomly, breaking the parser if it happens to one of these patterns.
        response = response.replace(/<,?m,?e,?t,?a[^>]+>/, '');	// /<meta[^>]+>/
        response = response.replace(/<,?s,?c,?r,?i,?p,?t,?>,?<,?!,?-,?-/g, '<script><![CDATA['); // /<script><!--/
        response = response.replace(/\/,?\/,?-,?-,?>,?<,?\/,?s,?c,?r,?i,?p,?t,?>/g, ']]></script>'); // /\/\/--><\/script>/

        // Load XML parser with response
        var parser = Components.classes['@mozilla.org/xmlextras/domparser;1'].createInstance(Components.interfaces.nsIDOMParser);
        var doc = parser.parseFromString(response, 'text/xml');

        // Create sandbox for eval context and add functions to it that Gmail expects to call
        var sb = Components.utils.Sandbox(url);
        var realArray = Array;
        sb.window = null;
        sb.top = {
            GG_iframeFn: function(junk, data) {
                // Takes the data chunk and pushes it on the data object stack (which is then extracted from the sandbox)
                if(!this.data) this.data = [];
                this.data.push(data);
            },
            _A: function() {
                // Creates array objects in the scope of the correct window (the chrome window in this case)
                var a;
                a = (arguments.length == 1 && typeof arguments[0] == 'number') ? [arguments[0]] : realArray.apply(null, arguments);
                return a;
            }
        };

        // Go through script tags and eval each in the sandbox
        var scripts = doc.getElementsByTagName('script');
        for(var i = 0; i < scripts.length; i++) {
	    try{
            Components.utils.evalInSandbox(scripts[i].textContent, sb);
	    }catch(e){
		// TODO Gmail seems to inject ',' into the script at random points, often causing a break. This error catching reduces the impact; but it's still fatal if it happens within the <SCRIPT><!-- (as replaced above), as our regexp can no longer purify the HTML. Do we have to protect the regexp with multiple comma tests?
	    }
        }

        if(typeof sb.top.data != 'object' ) throw new Error('result not an object');

        return glDataObject.createFromData(sb.top.data, gmail);
    },

    /**
     * Pass in the request response data object or the VIEW_DATA variable, and it
     * generates a glDataObject object.
     * @function {static glDataObject} ?
     * @param {Object} data - The data to generate the glDataObject object from
	 * @param {glGmail} gmail -
     * @return The generated glDataObject object
     */
    createFromData: function(data, gmail) {
        glLogger.log("Creating glDataObject object from data...", 'glDataObject');

        var viewDataObject = new glDataObject(gmail);

        for(var i = 0; i < data.length; i++) {
            if(!data[i] || !data[i].length) continue;

            var chunk = data[i];
            var chunkType = chunk[0];

            // Extract important chunks from the response data and store them in
            // the _rawData property, skipping unimportant ones to reduce memory
            // usage
            switch(chunkType) {
                case glDataObject.ADDITIONAL_ACCOUNTS_TAG:
                case glDataObject.LABEL_DATA_TAG:
                case glDataObject.LOGIN_ACTIVITY_TAG:
                case glDataObject.QUERY_DATA_TAG:
                case glDataObject.ACTIVITY_RESPONSE_TAG:
                    viewDataObject._rawData[chunkType] = chunk;
                    break;
                case glDataObject.THREAD_DATA_TAG:
                    // Can be multiple THREAD_DATA_TAG chunks, so they need to be processed differently
                    if(viewDataObject._rawData[chunkType] == undefined) viewDataObject._rawData[chunkType] = [];
                    viewDataObject._rawData[chunkType].push(chunk);
                    break;
                case glDataObject.THREAD_MESSAGE_DATA_TAG:
                    // Can be multiple THREAD_MESSAGE_DATA_TAG chunks, so they need to be processed differently
                    if(viewDataObject._rawData[chunkType] == undefined) viewDataObject._rawData[chunkType] = [];
                    viewDataObject._rawData[chunkType].push(chunk);
                    break;
                default:
                    // Do nothing
                    break;
            }
        }

        return viewDataObject;
    },
    //}
});
