/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Provides interfaces to the data that Gmail has. Label data, unread counts,
 * other e-mail addresses a user can send from, the thread id of the last sent
 * message, thread data for the threads in the current view, contacts. Basically
 * anything Gmail knows or can find out.
 *
 * @class ?
 */
/* Internal API Developers:
 * The API is structured in three parts: general environment data, special events
 * and event data, and traffic watcher data. General environment data is handled
 * by update and updateFromData and stored in _environmentData. It consists of
 * anything that is either constant, or doesn't need to be updated reqularly (yes
 * it's vague, but the other two parts make it easier to understand). Typically
 * it is accessed by calling something like getUserLabels or something. Special
 * events and event data consist of things like label name changes, sent e-mails,
 * and other like things. The way to tell if it falls in this category is to ask
 * whether the best way to convey that you have the information is by an event
 * and whether you're probably not going to save the information. This data is
 * always conveyed by event and rarely stored or cached. The final part, traffic
 * watcher data, consists of data that the traffic watcher grabs out of Gmail's
 * traffic and that is cached by glEnvironment. This includes things like thread
 * list data, conversation data that Gmail requests for improved performance, and
 * other things. The API for this can vary, but it is commonly accessed through
 * a general function like getThreadListData(). This way if there's a cache miss,
 * the function can still be ablt handle the data retrieval differently. Note that
 * just because the traffic watcher is being used to obtain the data doesn't mean
 * that the data can't be exposed through other parts of the API, like the action
 * detection.
 */
var glEnvironment = glBase.extend({
    // Properties {
    /**
     * The last environment data update date
     * @property {Date} ?
     */
    lastUpdate: null,

    /**
     * glGmail object
     * @property {glGmail} ?
     */
    _glGmail: null,

    /**
     * Stored environment data
     * @property {Object} ?
     */
    _environmentData: null,

    /**
     * The {@link glTrafficWatcher} instance. It will be null if it is not connected.
     * @property {glTrafficWatcher} ?
     */
    trafficWatcher: null,

    /**
    * The glThreadCache instance. 
    * @property {glThreadCache} ?
    */
    threadListDataCache: null,
    
    /** @property {Boolean} ? Refresh at the next request */
    _requestedRefresh: null,
    
    /** @property {glMultiToOne} ? */
    _multiToOne: null,
    
    /** @property {String} ? */
    lastSentThreadId: null,

    /**
     * Hash containing cached thread list data. The key is the URL that the request
     * was made to and the value is a hash containing the data (key "data"), and
     * a boolean flag (key "doDelete") that will trigger a delete of the data
     * the next time the view changes.
     * @property {Object} ?
     */
    //_threadListDataCache: null,	// TODO remove
    //}

    // Methods {
    /**
     * Connects the environment to the given glGmail object's page.
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    connect: function(gmail) {
        this._glGmail = gmail;
        this._environmentData = {};
        //this._threadListDataCache = {};
	this._multiToOne = new glMultiToOne(1500);


        // Update from view data
        var viewData = glDataObject.createFromData(gmail.page.getViewData(), this._glGmail);
        this.updateFromData(viewData, true);


        if( this._glGmail.plugins.isComponentEnabled('trafficWatcher') ){
            this._startTrafficWatcher();
        }
	
	this.threadListDataCache = new glThreadListDataCache(this._glGmail, viewData)

        this._connectToPageWrapper();
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    shutdown: function() {
        if(this.trafficWatcher) {
            this.trafficWatcher.shutdown();
            this.trafficWatcher = null;
        }

        this._environmentData = null;

        this._glGmail = null;

        this._removeAllListeners();
    },

    // Environment Data Related Methods {
    /**
     * Performs a simple synchronous request to repopulate the environment data.
     * @function {public} ?
     */
    update: function() {
        var request = glRequest.createSimpleRequest(this._glGmail);

        request.run(false); // Run synchronously
        if(request.errors.length) {
            glLogger.error("Update request had errors:\n"+glLogger.dump(request.errors, true), 'glEnvironment');
        }

        this.updateFromData(request.data);
    },

    /**
     * Updates the environment data store with the given data.
     * If the data is different, it dispatches an ENVIRONMENT_CHANGED event.
     * @function ?
     * @param {glDataObject} data - A glDataObject object containing environment data
     * @param {Boolean} full - If true, always include complete analysis
     */
    updateFromData: function(data, full) {
        glLogger.log('Updating environment from data @ '+(new Date()), 'glEnvironment');

        this.lastUpdate = new Date();

        if(this.hasEventListener(glEnvironment.ENVIRONMENT_CHANGED) || full) {
            // Someone is listening for ENVIRONMENT_CHANGED, so we need to actually determine if the environment changed
            var changed = [];

            var compare = {
                additionalAccounts: data.getAdditionalAccounts(),
                userLabels: data.getUserLabels(),
                internalLabels: data.getInternalLabels()
            };
            
            for(var set in compare) {
                if(compare[set] == null) continue;

                var currentData = this._environmentData[set];
                var newData = compare[set];

                // Compare currentData and newData - will probably need to be changed at some point to make it more generic, but it works for now
                var isNew = false;
                if(currentData && currentData.length == newData.length) {
                    for(var i = 0; i < newData.length; i++) {
                        for(var prop in newData[i]) {
                            if(currentData[i][prop] != newData[i][prop]) {
                                isNew = true;
                                break;
                            }
                        }
                        if(isNew) break;
                    }
                } else {
                    isNew = true;
                }
    
                if(isNew) {
                    changed.push(set);
                    this._environmentData[set] = compare[set];
                }
            }

            if(changed.length) {
                this.dispatchEvent({name: glEnvironment.ENVIRONMENT_CHANGED, changed: changed});
            }
        } else {
            // No one listening for ENVIRONMENT_CHANGED so don't bother doing calculations for it
            this._environmentData['additionalAccounts'] = data.getAdditionalAccounts();
            this._environmentData['userLabels'] = data.getUserLabels();
            this._environmentData['internalLabels'] = data.getInternalLabels();
        }
    },

    /**
     * Returns the account for the current environment.
     * @function {String} ?
     * @return The e-mail address of the active account
     */
    getAccount: function() {
        if(typeof this._environmentData['account'] == 'undefined') {
            this._environmentData['account'] = this._glGmail.page.getGlobal(10);
        }
        return this._environmentData['account'];
    },
    
    /**
    * Return array of user-created labels
    * @function {} ?
    * @return Array of user-created labels
    */
    getUserLabels: function(){
    	return this._environmentData.userLabels;
    },
    

    /**
     * The IK value is some type of cookie value that Gmail requires in requests
     * to the server.
     * @function {String} ?
     * @return The IK value
     */
    getIKValue: function() {
        if(typeof this._environmentData['ikValue'] == 'undefined') {
            this._environmentData['ikValue'] = this._glGmail.page.getGlobal(9);
        }
        return this._environmentData['ikValue'];
    },

    /**
     * Returns the URL that requests should be made to. Used by {@link glRequest}
     * to build the request URL.
     * @function {String} ?
     * @return The request URL
     */
    getRequestURL: function() {
        if(typeof this._environmentData['requestURL'] == 'undefined') {
            this._environmentData['requestURL'] = this._glGmail.page.getGlobal(7);
        }
        return this._environmentData['requestURL'];
    },
    //}

    // glPageWrapper Related Methods {
    /**
     * Connects important event listeners to the page wrapper object.
     * @function ?
     */
    _connectToPageWrapper: function() {
        this._glGmail.page.addEventListener(glPageWrapper.VIEW_CHANGED, this._viewChanged, this);
    },

    /**
     * glPageWrapper view change callback
     * @function ?
     * @param {Event} evt - The {@link glPageWrapper.VIEW_CHANGED} event
     */
    _viewChanged: function(evt) {
        // GC _threadListDataCache, flagging new cached items and deleting flagged items
		// Garbage collect old items - if 10 mins since updated, delete. 
	var d = (new Date()).valueOf() - (1000*60*10);
        for(var key in this._threadListDataCache) {
		if( this._threadListDataCache[key].created < d && !this._threadListDataCache[key].mostRecent ){
		        delete this._threadListDataCache[key];
		}
        }
        
        if( this._requestedRefresh ){
        	// If can refresh UI, do so
        	if( this._glGmail.page.getActiveViewType()=="tl" ){
        		this._requestRefresh();
        	}
        }
    },
    //}

	/**
	* Request a refresh at the earliest opportunity 
	* The Ajax requests - especially in the popup - do not cause Gmail to update; so we must ask it to otherwise users will doubt the integrity.
	* @function {} ?
	*/
	requestRefresh: function(){
		this._multiToOne.process("requestRefresh", glBase.bind(this, this._requestRefresh));	
	},
	
	/**
	* Actually attempt the refresh
	* @function {} ?
	*/
	_requestRefresh: function(){
		if( this._glGmail.page.getActiveViewType()=="tl" ){
			this._requestedRefresh = false;
			var actionsHeader = this._glGmail.getCurrentViewUIObject().actionsHeader;
			var el = actionsHeader.getButton(glActionsHeader.REFRESH_BUTTON);
			if( el ){
				glUIWidget.simulateMouseEvent(el, 'mousedown');
				glUIWidget.simulateMouseEvent(el, 'mouseup');
				glUIWidget.simulateMouseEvent(el, 'click');
			}
			this._requestedRefresh = false;
		}else{
			this._requestedRefresh = true;
		}
	},

    // Traffic Watcher Related Methods {
    /**
     * Checks the internal thread list data cache to see if it has data matching
     * the request and returns it.
     * @function {Array} ?
     * @param {String} viewHash - The hash portion of the URL for the current view ('#search/Apple+computers' for example)
     * @param {int} startOffset - The start offset of the thread data
     * @param {int} endOffset - The end offset of the thread data
     * @return The threads that it found - empty array if none found
     */

    /**
     * Initializes a traffic watcher and starts it after connecting important
     * event listeners.
     * @function ?
     */
    _startTrafficWatcher: function() {   
        if(this.trafficWatcher) return;

        this.trafficWatcher = new glTrafficWatcher();
        this.trafficWatcher.addEventListener(glTrafficWatcher.SEND_MESSAGE_REQUEST_COMPLETED, this._handleSendMessageRequest, this);
	this.trafficWatcher.addEventListener(glTrafficWatcher.NEW_MESSAGE_REQUEST, this._handleNewMessageRequest, this);
        this.trafficWatcher.addEventListener(glTrafficWatcher.ACTION_REQUEST, this._handleActionRequest, this);
        //this.trafficWatcher.addEventListener(glTrafficWatcher.THREAD_LIST_REQUEST_COMPLETED, this._handleThreadListRequest, this);
        this.trafficWatcher.connect(this._glGmail);
    },

    /**
     * Event handler for {@link glTrafficWatcher.SEND_MESSAGE_REQUEST_RECEIVED} events.
     * Dispatches the EMAIL_SENT event with the thread id of the sent e-mail if
     * it is in the response. Also updates the internal environment data from
     * the response.
     * @function ?
     * @param {Event} evt - The SEND_MESSAGE_REQUEST_RECEIVED event
     */
    _handleSendMessageRequest: function(evt) {
        // Is there a thread id for the sent e-mail?
        var actionResponse = evt.responseData.getActionResponseData();
        if(actionResponse && actionResponse.affectedThread) {
	    this.lastSentThreadId = actionResponse.affectedThread;
            // Dispatch EMAIL_SENT event with the thread id
            this.dispatchEvent({name: glEnvironment.EMAIL_SENT, threadId: actionResponse.affectedThread});
        }


        // Update internal environment data from response
        this.updateFromData(evt.responseData);
    },
    
    _handleNewMessageRequest: function(evt){
	// Extract the thread id:
	if( this._glGmail.fb ) this._glGmail.fb.log('new message:'); // DEBUG
	if( this._glGmail.fb ) this._glGmail.fb.log(evt); // DEBUG
	var actionResponse = evt.responseData.getActionResponseData();
	if( this._glGmail.fb ) this._glGmail.fb.log(actionResponse); // DEBUG
	
	var m = /[?&]t=([^&]+)/.exec(evt.url);
	if( m ){
	    var threadId = m[1];
	    this.dispatchEvent({name: glEnvironment.NEW_MESSAGE, threadId: threadId});
	}
    },

    /**
     * Event handler for {@link glTrafficWatcher.ACTION_REQUEST} events. Dispatches
     * the ACTION_PERFORMED event with the url and the post data.
     * @function ?
     * @param {Event} evt - The ACTION_REQUEST event
     */
    _handleActionRequest: function(evt) {
	var m = evt.url.match(/[?&]act=([^&]+?)(&|$)/);	// TODO should type-finding logic be embedded into TrafficWatcher? Or Action Manager? Or here? Ok for now - problem arises if other code needs similarly logic.
	if( m ){
		var act = m[1];
		if( act.substr(0,3)=="cc_" ){
			// Add a label
			var label = unescape(act.substr(3));
			this._environmentData.userLabels.push({name:label});
	                this.dispatchEvent({name: glEnvironment.ENVIRONMENT_CHANGED});
		}else if( act.substr(0,3)=="nc_" ){
			// Rename a label
			var label = unescape(act.substr(3)).split("^");	// Format: [oldlabel, newlabel]
			for( var i = 0; i < this._environmentData.userLabels.length; i++ ){
				if( this._environmentData.userLabels[i].name==label[0] ){
					this._environmentData.userLabels[i].name = label[1];
					break;
				}
			}
	                this.dispatchEvent({name: glEnvironment.ENVIRONMENT_CHANGED});
		}else if( act.substr(0,3)=="dc_" ){
			// Delete a label
			var label = unescape(act.substr(3));
			for( var i = 0; i < this._environmentData.userLabels.length; i++ ){
				if( this._environmentData.userLabels[i].name==label[0] ){
					this._environmentData.userLabels.splice(i,1);
					break;
				}
			}
	                this.dispatchEvent({name: glEnvironment.ENVIRONMENT_CHANGED});
		}
	}
    
        this.dispatchEvent({name: glEnvironment.ACTION_PERFORMED, url: evt.url, postData: evt.postData});
        
    },

    /**
     * Event handler for {@link glTrafficWatcher.THREAD_LIST_REQUEST_RECEIVED}
     * events. Adds the response data to the thread list cache.
     * @function ?
     * @param {Event} evt - The THREAD_LIST_REQUEST_RECEIVED event
     */
	/*
    _handleThreadListRequest: function(evt) {
		for( key in this._threadListDataCache ) this._threadListDataCache[key].mostRecent = false;
		if( this._threadListDataCache[evt.url] ) delete this._threadListDataCache[evt.url];
        this._threadListDataCache[evt.url] = {data: evt.responseData, doDelete: false, mostRecent: true, created: (new Date()).valueOf()};
    },
	*/
	
    //}
    //}
}, {
    // Static Properties {
    
    /**
     * Dispatched when an e-mail arrives
     * @event ?
     */
    NEW_MESSAGE: 'new_message',
    
    /**
     * Dispatched when an e-mail is sent and includes the thread id of the sent e-mail
     * @event ?
     */
    EMAIL_SENT: 'email_sent',

    /**
     * Dispatched when an action is performed with the url and the post data
     * @event ?
     */
    ACTION_PERFORMED: 'action_performed',

    /**
     * Dispatched when environment data is changed in an update
     * @event ?
     */
    ENVIRONMENT_CHANGED: 'environment_changed',
    //}
});
glEnvironment.implement(glEventDispatcher); // Add event dispatching support
