/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * The base object for a given browser tab with Gmail open. It coordinates all of
 * the gmail components and provides references to the document and windows for
 * the Gmail page.
 *
 * @class ?
 */
var glGmail = glBase.extend({
    // Properties {
    /**
     * The document object for the page
     * @property {Document} ?
     */
    doc: null,

    /**
     * A reference to the canvas iframe content document
     * @property {Document} ?
     */
    canvasDocument: null,

    /**
     * The window object for the page
     * @property {Window} ?
     */
    window: null,

    /**
     * The {@link glPageWrapper} object for the page
     * @property {glPageWrapper} ?
     */
    page: null,

    /**
     * The {@link glEnvironment} object for the page
     * @property {glEnvironment} ?
     */
    environment: null,

    /**
     * The {@link glActionManager} object for the page
     * @property {glActionManager} ?
     */
    actions: null,


    /**
     * A hash of all the UI objects
     * @property {Object} ?
     */
    _ui: null,
    
    /**
     * Plugins loaded for GTDInbox
     * @property {glPlugins} ?
     */
    plugins: null,
    
    
    /**
     * Account data object for the current page
     * @property {glAccountData} ?
     */
    accountData: null,
    
    //}

    /**
     * The constructor.
     * @constructor ?
     * @param {Document} doc - The document element for the page to attach to
     * @param {glAccountData} accountData - The user's account for the current page
     * @param {glPlugins} plugins - The glPlugins object
     */
    constructor: function(doc, accountData, plugins) {
        glLogger.log('glGmail object constructing', 'glGmail');

	
        // Initialize instance variables
        this.doc = doc;
        this.window = doc.defaultView;
        this._ui = {};
        this.plugins = plugins;
        this.accountData = accountData;
        
	

        // Set up page wrapper
	
        this.page = new glPageWrapper();
        this.page.connect(this);
	
	
        // Check version and if it's not 2, throw an exception
        if(!/ui=2/.test(this.page.getGlobal(11))) { // Global 11 is the url for the javascript code for gmail
            glLogger.log('Global 11: '+this.page.getGlobal(11));
            this.page.disconnect();
            throw new Error("glGmail only works with version 2 of GMail");
        }
	
	
	// Set up environment
        this.environment = new glEnvironment();
        this.environment.connect(this);
    
    
        // Set up action manger
        this.actions = new glActionManager();
        this.actions.connect(this);
	
		
	// Initiate Gmail events: (Nb glGmail will always fire this before plugins)
	if( this.page.isUIReady() ){
	    this._baseUIReady();
	}else{
	    this.page.addEventListener(glPageWrapper.UI_READY, this._baseUIReady, this);
	}
	this.page.addEventListener(glPageWrapper.VIEW_CHANGED, this._baseViewChanged, this);
	if( this.page.hasFiredFirstViewChange() ){
	    // We missed our first load; so fire it manually
	    this._baseViewChanged();
	}

    },

    // Methods {
    
    
    /**
     * Called on page unload, it cleans up everything
     * @function ?
     */
    unload: function() {
        for(var element in this._ui) {
            this._ui[element].disconnect();
        }

        this.environment.shutdown();

        this.canvasDocument = null;
        this.window = null;
        this.doc = null;
    },

    
    

    /**
     * Initializes a {@link glSearch} object and returns it
     * @function {glSearch} ?
     * @return The searcher object
     */
    getSearcher: function() {
        return new glSearch(this);
    },

    /**
     * Returns the given UI object after initializing it (if necessary).
     * For example, if you wanted a glActionsDropdown object, you would call
     * <code>getUIObject('actionsDropdown')</code>, dropping the gl and turning
     * the 'A' to an 'a'. If it has already been initialized/created already, it
     * returns the cached one (when possible).
     * @function {Object} ?
     * @param {String} type - The UI object to return
     * @return The initialized UI object
     */
    getUIObject: function(type) {
        glLogger.log('Getting '+type+'...', 'glGmail');

        var object = null;
        if(!this._ui[type]) {
            // Not in cache so create and cache (or return if not supposed to cache)

            // Add gl to the begining of the type and raise the first character up
            // mainView -> glMainView
            theClass = 'gl'+type.substr(0,1).toUpperCase()+type.substr(1);

            glLogger.log('Creating object: "'+theClass+'"', 'glGmail');

            try {
                object = new window[theClass]();
            } catch(e) {
                // Use glLogger.error because it gives a stack trace
                glLogger.error("Failure instantiating "+theClass+" ui object:\n"+e.message, 'glGmail');
            }

            object.connect(this);

            // Is the ui object cachable - return if not
            if(object.isCacheable()) {
                // Cache the instantiated object
                this._ui[type] = object;
            }
        } else {
            // Get object from cache
            object = this._ui[type];
        }

        return object;
    },

    /**
     * Returns the UI object for the current view, initializing it if it hasn't
     * already been loaded.
     * @function {glUIView} ?
     * @return The view UI object
     */
    getCurrentViewUIObject: function() {
        if(!this.page.isUIReady()) return null;

        // In the cache, so return
        if(typeof this._ui['currentViewObject'] != 'undefined') {
            return this._ui['currentViewObject'];
        }

        // Get the UI object type for a call to getUIObject
        var uiObject = '';
        switch(this.page.getActiveViewType()) {
            case 'tl': uiObject = 'threadListView'; break;
            case 'cv': uiObject = 'conversationView'; break;
            case 'co': uiObject = 'composeView'; break;
            case 'ct': uiObject = 'contactsView'; break;
        }

        if(uiObject == '') {
            return null;
        } else {
            // Use getUIObject so extending classes can override the view objects without overriding this method
            this._ui['currentViewObject'] = this.getUIObject(uiObject);
        }

        return this._ui['currentViewObject'];
    },

    /**
     * The UI_READY callback for the glGmail class. Currently just populates
     * the canvasDocument property.
     * @function ?
     * @param {Event} evt - The glPageWrapper.UI_READY event
     */
    _baseUIReady: function(evt) {
        // UI ready, so create reference to canvas document
        glLogger.log('Base UI Ready, so retrieve canvas document', 'glGmail');
        this.canvasDocument = this.doc.getElementById("canvas_frame").contentDocument
	this.lightbox = this.getUIObject('lightbox');
	glLogger.log('Base UI Ready... retrieved canvas document', 'glGmail');
	
	glLogger.outputWarningStack("glGmail._baseUIReady", this);
	
	// If not Installed, install now
	// Install routine will handle establishing a server account and showing welcome screen.
	// It will also disable GTDInbox if the user does not wish to use it.
	// TODO 
	
	// Otherwise, Check if last password attempted on a server sync was wrong.
	// If so, show the Change Password screen; which takes care of updating the password, getting the latest settings, merging them, and then asking the user to refresh Gmail.
	// TODO
	
    },

    /**
     * glPageWrapper view change callback. Currently just disconects and invalidates
     * the currentViewObject cache if it is populated.
     * @function ?
     * @param {Event} evt - The glPageWrapper.VIEW_CHANGED event
     */
    _baseViewChanged: function(evt) {
        // Disconnect and invalidate currentViewObject cache if populated
	
	glLogger.log("base view changed", "glGmail");
	
        if(typeof this._ui['currentViewObject'] != 'undefined') {
            if(this._ui['currentViewObject']) this._ui['currentViewObject'].disconnect();
            delete this._ui['currentViewObject'];
        }
	
    },
    //}
},{

    // Static Methods {
    /**
     * Checks if the given URL is for the main gmail page and returns true if it
     * is.
     * @function {static Boolean} ?
     * @param {String} url - The url to check
     * @return Whether or not it's the main gmail page URL
     */
    isMainGmailPageUrl: function(url) {
        // Check that base portion is okay
        if(!(new RegExp('^http(s?)://mail.google.com/mail/')).test(url) && !(new RegExp('^http(s?)://mail.google.com/a/[^/]+/')).test(url)) {
            return false;
        }

        // Does it have a view url parameter (common for popout pages)
        if(/[?&]view=/.test(url)) {
            return false;
        }

        return true;
    },
    
    addCSS: function(doc, href){
	    var elLink = doc.createElement("LINK");
	    elLink.type = "text/css";
	    elLink.href = href;	
	    elLink.rel = "stylesheet";
	    doc.getElementsByTagName("HEAD")[0].appendChild( elLink );
    },

    
    //}
});
