/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Instantiates, maintains reference to, and provides helper functions for GTDInbox plugins (including gtd).
 * Executes when glGmail loads. 
 * glGmail is the only code instantiated when Gmail opens; that loads a user's account data, which defines which plugins they wish to use.
 *
 * @class ?
 */

var glPlugins = glBase.extend({
   
   
    // Properties {
    
    /**
     * Reference to the glAccountData object for the page
     * @property {glAccountData} ?
     */
    _accountData: null,
    
    /**
     * Hash of instantiated plugins
     * @property {Object{}} ?
     */
    _plugins: null,
    
    /**
     * @property {mozIJSSubScriptLoader} ?
     */
    _scriptLoader: null,
    
    //}
    
    // Methods {
    
    
    /**
     * The constructor.
     * Instantiate each glPluginBase-derived plugin specified in the settings (and not disabled)
     * @constructor ?
     * @param {glAccountData} accountData - 
     */
    constructor: function(accountData) {
        glLogger.log('constructing glPlugins', 'glPlugins');

        this._accountData = accountData;
        this._plugins = {};
        
        // Retrieve the plugin details:
        var plugins = this._accountData.getPref('plugins');
        
        
        // Load each plugin
        for( id in plugins ){
            var plugin = plugins[id];
            if( plugin && !this._plugins[id] && !plugin.disabled ){
                try{
                    this.loadChromeFile(plugin.path);
                    this._plugins[id] = new window[plugin.className]();
                }catch(e){
                    glLogger.warn('Plugin '+id+' failed to load: ' + e, 'glPlugins.constructor');
                }
            }
        }
    },
    
    /**
     * Instruct each plugin to load into gmail
     * @function {} ?
     * @param {glGmail} gmail
     */
    loadPluginsForGmail: function(gmail){
        glLogger.log('load plugins for gmail', 'glPlugins');
        
        for( id in this._plugins ){
            try{
                this._plugins[id].loadForGmail(gmail);
            }catch(e){
                glLogger.warn('Plugin '+id+' failed to load for gmail: ' + e, 'glPlugins.loadPluginsForGmail');
            }
        }
        
    },
    
    
    /**
     * Helper function to load the designated file into memory.
     * Must be prefixed with "chrome://"
     * @function {} ?
     * @param {String} filename - The chrome file to load
     */
    loadChromeFile: function(filename){
        if( !this._scriptLoader ){
            this._scriptLoader = Components.classes["@mozilla.org/moz/jssubscript-loader;1"].getService(Components.interfaces.mozIJSSubScriptLoader);
        }
        
        // TODO do we need a static array/hash to record files opened per browsing session? i.e. don't read files on each page load
        
        if( filename.indexOf("chrome://")!=0 ) return;
        
        this._scriptLoader.loadSubScript(filename);
    },
    
    /**
     * Returns an array of plugin instances
     * @function {glPluginBase[]} ?
     * @return Array of plugin instances
     */
    getLoadedPlugins: function(){
        var plugins = [];
        for( id in this._plugins ){
            plugins.push( this._plugins[id] );
        }
        return plugins;
    },
    
    /**
     * Returns whether or not the given component should be enabled
     * @function {Boolean} ?
     * @param {String} component - The component name
     * @return Whether or not the component should be enabled
     */
    isComponentEnabled: function(component) {
        
        if( !this._requiredComponents ){
            this._requiredComponents = {};
            for( id in this._plugins ){
                var req = this._plugins[id].getRequiredComponents();
                for( var i = 0; i < req.length; i++ ){
                    this._requiredComponents[req[i]] = true;
                }
            }
        }
        
        if( this._requiredComponents[component] ){
            return true;
        }else{
            return false;
        }
        
    },
  
    /**
     * Call unload on all plugins
     * @function {} ?
     */
    unload: function(){
        if( this._plugins ){
            for( p in this._plugins ){
                try{
                    this._plugins[p].unload();
                }catch(e){
                    glLogger.log('Plugin unload failed for ' + p + ':' + e.toString(), 'glPlugins');
                }
            }
            this._plugins = null;
        }
        
        this._accountData = null;
    },
    
    // }
    
},{
    
});