/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Thread data wrapper class. Provides an easy interface to the thread data and
 * common actions on threads.
 *
 * @class ?
 */

// TODO There are essentially two kinds of thread objects - those loaded by threadlist, and those loaded by conversation (both ajax calls). Should we be sub-classing this object to make that more logical?

var glThread = glBase.extend({
    // Properties {

	/** @property {Boolean} ? Used to know if loading */
	_loaded: false,

	/**
     * glGmail object
     * @property {glGmail} ?
     */
    _gmail: null,

    /**
     * glThreadDetails object
     * @property {glThreadDetails} ?
     */
    _threadDetails: null,

    /**
     * The thread's given thread id
     * @property {String} ?
     */
    threadId: null,

    /**
     * An array containing all the thread's labels (internal and user created)
     * @property {String[]} ?
     */
    labels: null,

    /**
     * The subject of the message that Gmail displays in the view (so it could be truncated)
     * @property {String} ?
     */
    subjectSnippet: null,

    /**
     * The date string for when the e-mail was sent (or received?). If it's
     * necessary, a parser could be written to turn this into a real Date object.
     * @property {String} ?
     */
    date: null,
    
    /** @property {String} - The date rendered in Gmail's list */
    dateSnippet: null,
    
    /** @property {String} - Comma seperated list of files - beware Gmail does not escape commas */
    attachmentsCSV: null,
        
    /** @property {String[]} - Array of email addresses. Format: [{name:,email:}] */
    knownContacts: null,

    /** @property {int} - The Gmail defined timestamp */
    timestamp: null,
    
    /** @property {int} - The timestamp that is compatible with the JS Date object */
    timestampJS: null,
    
    //}


	// Optional Properties {
	// Note: The following variables are not always available. It depends whether it was a thread-list (tb) load, or a conversation (ms) load.

    /**
     * Whether or not the thread has been read
     * @property {Boolean} ?
     */
    isRead: false,

    /**
     * Whether the user has starred the thread
     * @property {Boolean} ?
     */
    isStarred: false,

    /**
     * Whether the thread is deleted or not
     * @property {Boolean} ?
     */
    isDeleted: false,

    /** @property {String} - The Gmail preview line */
    bodySnippet: null, 

    /** @property {String} - The Gmail name line */
    contactsSnippet: null,

	// }

    

    // Methods {

    /**
    * The constructor.
    * @constructor ?
    */
    constructor: function(gmail){
        this._gmail = gmail;
    },
	
    /**
     * Converts the thread to a string representation
     * @function {String} ?
     * @return The string representation
     */
    toString: function() {
        return this.threadId+": "+this.subjectSnippet+" - "+this.date;
    },
    
    /**
    * Find if this thread has a particular label
    * @function {boolean} ?
    * @param {String} labelname - The label to test for
    * @return True if the label was found on the thread
    */
    hasLabel: function(labelname){
    	return (this.labels.indexOf(labelname)>-1);
    },

    /**
     * Return true if the term is found within the subject or the body of the thread
     * @function {String} ?
     * @param {String} term - The search term to match exactly
     * @return True if found
     */
    hasSearchTerm: function(term){
        if( this.subjectSnippet.indexOf(term)>-1 ) return true;
        if( this.bodySnippet.indexOf(term)>-1 ) return true;
        if( this._threadDetails ){
            return this._threadDetails.hasSearchTerm(term);
        }
    },

    /**
    * Return true if tested contact is related to this thread
    * @function {Boolean} ?
    * @param {optional String[]} contacts - An array of names or email addresses - will try to match any one of them. For instance, you might try an email address followed by their name.
    * @param {optional Boolean} isSender - If true, only return if contact sent a message on this thread
    * @param {optional Boolean} isRecipient - If true, only return if contact is a recipient a message on this thread
    * @return true if contact found
    */
    hasContact: function(contacts, isSender, isRecipient){
        if( typeof contacts=="string" ) contacts = [contacts];
        if( (typeof isSender=='undefined') && (typeof isRecipient=='undefined') ){    // Only test general items if not requiring indepth knowledge of senders/recipients
            for( var i = 0; i < contacts.length; i++ ){
                if( glBase.testEmailAddress(contacts[i], true) ){
                    // It's an email
                    if( m = /^\"(.+)\"$/.exec(contacts[i]) ) contacts[i] = m[1];    // Strip quotes 
                    for( var j = 0; j < this.knownContacts.length; j++ ){
                        if( this.knownContacts[j].email==contacts[i] ) return true;
                    }
                }else{
                    // It's a name
                    for( var j = 0; j < this.knownContacts.length; j++ ){
                        if( glThread._hasContact_checkName(contacts[i], this.knownContacts[j].name) ) return true;
                    }
                }
            }
        }
        if( this._threadDetails ){
            return this._threadDetails.hasContact(contacts, isSender, isRecipient);
        }
        return false;
    },

    /**
    * True if account owner sent the email to themselves
    * @function {Boolean} ?
    * @return True if account owner sent the email to themselves
    */
    isPersonal: function(){
        if( this.knownContacts.length==1 ){
            if( this.knownContacts[0].email==this._gmail.environment.getAccount() ){
                return true;
            }
            // TODO Environment doesn't load cfs from VIEW_DATA when it constructs; and it's updateFromEvent is called at no other time. Need a way to force it to load environment variables.
        }else{
            return false;
        }
    },

    /**
    * Retrieve the thread details object, which contains all messages (and subsequently more accurate date/file info)
    * If glThreadDetails object doesn't exist, create it, and load it for this thread.
    * It's possible the threadDetails will be returned without full data loaded. Calling function is advised to attach a glThreadDetails.LOADED listener and respond to guarentee accurate information.
    * @function {glThreadDetails} ?
    * @param {boolean} doNotLoad - If true, do not try to retrieve them from the server 
    */
    getThreadDetails: function(doNotLoad){
        if( !this._threadDetails && !doNotLoad ){
            this._threadDetails = new glThreadDetails(this._gmail);
            this._threadDetails.addEventListener(glThreadDetails.LOADED, this._threadDetailsUpdate, this);
            this._threadDetails.load(this.threadId);
            this._loading = true;
        }
        return this._threadDetails;
    },
    
    /**
    * Notification that the this._threadDetails object updated. Propogate.
    * @function ?
    */
    _threadDetailsUpdate: function(event){
        this._loading = false;
        this.dispatchEvent({name:glThread.LOADED_DETAILS});
    },

    /**
    * Return true if this thread, or one of its messages, matches the passed ID
    * It's required because Gmail does not always use consistent IDs when referring to a thread.
    * @function {Boolean} ?
    * @param {String} id - The ID
    * @return True, if id matches something in this thread
    */
    hasId: function(id){
        if( id==this.threadId ) return true;
        if( this._threadDetails ){
            for( var i = 0; i < this._threadDetails.messages.length; i++ ){
                if( this._threadDetails.messages[i].messageId==id ) return true;
            }
        }
        return false;
    },
    
    /**
    * Each message has its own ID, and Gmail does not always use the most recent as the .threadId
    * It appears the archive function fails unless the most recent ID is used.
    * @function {String} ?
    * @return The threadId from the last message, or the thread id
    */
    getRecentId: function(){
        if( this._threadDetails && this._threadDetails.messages ){
            var msg = this._threadDetails.messages[this._threadDetails.messages.length-1];
            return msg? msg.messageId : this.threadId;
        }else{
            return this.threadId;
        }
    },

    /**
     * Updates the internal data representation using the given event if it applies
     * @function {Boolean} ?
     * @param {Event} evt - A {@link glActionManager} action event
     * @return Whether the thread was updated/affected by the event
     */
    updateFromEvent: function(evt) {
        if(evt.name == glActionManager.THREAD_CHANGED) {
            if(evt.thread != this.threadId) return false;

            glLogger.log('Updating thread from event: '+this.threadId, 'glThread');
            switch(evt.action) {
                case glActionManager.ADD_LABEL:
                    if( this.labels.indexOf(evt.label)==-1 ) this.labels.push(evt.label);
                    break;
                case glActionManager.REMOVE_LABEL:
                    var index = this.labels.indexOf(evt.label);
                    if(index != -1) {
                        this.labels.splice(index, 1);
                    }
                    break;
                case glActionManager.STAR:
                    this.isStarred = true;
                    break;
                case glActionManager.UNSTAR:
                    this.isStarred = false;
                    break;
                case glActionManager.ARCHIVE:
                    var index = this.labels.indexOf('^i');
                    if(index != -1) {
                        this.labels.splice(index, 1);
                    }
                    break;
                case glActionManager.REPORT_SPAM:
                    if( this.labels.indexOf("^s")==-1 ) this.labels.push("^s");
                    break;
                case glActionManager.NOT_SPAM:
                    var index = this.labels.indexOf("^s");
                    if(index != -1) {
                        this.labels.splice(index, 1);
                    }
                    break;
                case glActionManager.DELETE:
                    this.isDeleted = true;
                    break;
            }
            return true;
        } else {
            return false;
        }
    },

    /**
    * Load the data for a given thread id, via ajax.
    * @function {} ?
    * @param {String} threadId - the id of the thread
    */
    _load: function(threadId){
        this.threadId = threadId;
        this._loading = true;
        var request = new glRequest(this._gmail, { 'view': 'cv', 'th':threadId, 'prf':'1', 'rt':'j', 'search':'query', 'q':'' });
        request.addEventListener(glRequest.COMPLETED, this._loadCallback, this);
        request.run();
    },
    
    /**
    * Fired when ajax data for a thread load returned
    * Parse the data (including details), and notify listeners that the data has changed
    * @function ?
    */
    _loadCallback: function(event){
        this._loading = false;
        var request = event.target;
        var data = request.data.getThreadData(true);
        var success = null;
        if( data ){
            success = true;
            for(p in data){
                if( data.hasOwnProperty(p) ){
                    this[p] = data[p];
                }
            }
            this._threadDetails = new glThreadDetails(this._gmail);
            var me = this;
            this._threadDetails.addEventListener(glThreadDetails.LOADED, function(){
                me.dispatchEvent({name:glThread.LOADED_DETAILS, thread:me});
            }, this);
            this._threadDetails.load(this.threadId, request.data);	// Load using current data object
        }else{
            success = false;
            this.threadId = null;
        }
        
        // Notify of update
        this.dispatchEvent({name: glThread.LOADED, thread:this, success:success});
    },
    
    /**
    * Find out if thread is loading.
    * If not, will need to listen for glThread.LOADED.
    * Does not distinguish between a full thread load, or loading the details. 
    * @function {Boolean} ?
    */
    isLoading: function(){
        return this._loading;
    },
	

    //}
},{
	
    // Static Properties {
    
    LOADED: 'loaded',
    
    LOADED_DETAILS: 'loaded_details',
    
    // }
	
    // Static Methods {
	
    /**
    * Create a glThread object and set it to load (via ajax) thread data
    * As this is async, upon completion the glThread object will fire a glThread.THREAD_DATA_UPDATE event
    */
    createFromId: function(threadId, gmail){        
        var thread = new glThread(gmail);
        thread._load(threadId);
        return thread;
    },
	
    /**
     * Creates and returns a glThread object populated from the given data.
     * The reason it is passed a pre-parsed object from {@link glDataObject} is
     * that I'd like to keep most of the code that can change in a few places (the
     * UI widgets being an exception for the sake of code beauty and readability).
     * I'm not really attached to that decision though, so the parsing may be
     * moved into here.
     * @function {static glThread} ?
     * @param {Object} threadData - Object containing exctracted thread data
	 * @param {glGmail} gmail - the glgmail page object
     * @return The created glThread obect
     */
    createFromData: function(threadData, gmail) {
        var thread = new glThread(gmail);

        // Poplulate created thread object
        var wantedProps = ['threadId', 'isRead', 'isStarred', 'labels', 'subjectSnippet', 'dateSnippet', 'date', 'attachmentsCSV', 'bodySnippet', 'contactsSnippet', 'knownContacts', 'timestamp'];
        for(var i = 0; i < wantedProps.length; i++) {
            thread[wantedProps[i]] = threadData[wantedProps[i]];
        }
        if( thread.timestamp ){
            // Convert to JS date format, which is 3 characters less
            var timestampJS = new String(thread.timestamp);
            timestampJS = timestampJS.substring(0, timestampJS.length-3);
            thread.timestampJS = parseInt(timestampJS);
        }
        

        return thread;
    },
    
    /**
     * TODO This is used by glThread and glThreadMessage. Can we give it a better (more logical) home? E.g. a parent class
     * Advanced, Gmail-esque test for a name against a name stored on a thread in regular format.
     * So, 'Bob' will match 'Bob Jones', but 'Bo' will not. Similarly, 'Pete' can match 'Pete Jones', but cannot match 'Peter Jones'. Handle instances with surrounding quotes differently
     * @function {Boolean} ?
     * @param {String} testName - The name to try. Can be in quotes, or partial.
     * @param {String} knownName - The name as it is stored on the thread
     * @return True if testName can be considered a match for knownName
     */
    _hasContact_checkName: function(testName, knownName){
        if( m = /^\"(.+)\"$/.exec(testName) ){
            // Surrounded by quotes
            if( m[1].indexOf(" ")>-1 ){
                // Contains a space, fall back to slightly less sophisticated test
                return (knownName.indexOf(m[1])>-1);
            }else{
                // Strip quotes
                testName = m[1];
            }
        }
        // Test names, observing spaces:
        var tokensName = testName.split(" ");
        var tokensKnown = knownName.split(" ");
        for( var k = 0; k < tokensName.length; k++ ){
            for( var l = 0; l < tokensKnown.length; l++ ){
                if( tokensName[k]==tokensKnown[l] ) return true;
            }
        }                    
    },
    
    //}
});
glThread.implement(glEventDispatcher); // Add event dispatching support