/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * File containing code that is injected into the gmail page when glPageWrapper
 * is connected.
 * @file content/js/gmail/injected-code.js
 */
(function() {

/* Internal API Developers:
 * You cannot use eval in any event listeners that will be triggered by glPageWrapper,
 * but I'm not sure exactly why - it just doesn't work.
 */

/**
 * Root object of all the injected code
 * @namespace ?
 */
var pageWrapper = {
    /**
     * Sets up event listeners for the special injected element to allow communication
     * and start gmonkey object loading. Called at the end of the file.
     * @function ?
     */
    init: function() {
        // Set up event listeners
        var elem = pageWrapper._getElem();
        elem.addEventListener('glGmailRunFunction', pageWrapper._runFunctionListener, false);

        // Start grease monkey load process
        pageWrapper.gmonkeyLoaded = false;
        pageWrapper._loadGreasemonkey();
    },

    /**
     * Returns the object at the given offset in the GLOBALS array.
     * @function {Object} ?
     * @param {int} offset - The offset in the GLOBALS array
     * @return The object at the given offset
     */
    getGlobal: function(offset) {
        return unsafeWindow.GLOBALS[parseInt(offset)];
    },

    /**
     * Returns the VIEW_DATA object.
     * @function {Object} ?
     * @return the VIEW_DATA object
     */
    getViewData: function() {
        return unsafeWindow.VIEW_DATA;
    },

    /**
     * Returns the active view type if the gmonkey object is available
     * @function {String} ?
     * @return The active view type
     */
    getActiveViewType: function() {
        if(!pageWrapper.gmonkeyLoaded) {
            return null;
        } else {
            return pageWrapper._gmonkey.getActiveViewType();
        }
    },

    // Grease monkey code {
    /**
     * Starts the loading process of the gmonkey object. Calls itself regularly
     * until the gmonkey object is available and passes it to
     * <code>_greasemonkeyLoaded</code> when it is.
     * @function ?
     */
    _loadGreasemonkey: function() {
        try {
            if(pageWrapper._gmonkey) return;
            if(!unsafeWindow.gmonkey) { // gmonkey not yet loaded so wait
                setTimeout(function() { pageWrapper._loadGreasemonkey(); }, 200);
                return;
            }

            // Needed because sometimes doesn't call the load callback on Linux
        
            if(!pageWrapper._calledGmonkeyLoad) {
                pageWrapper._calledGmonkeyLoad = true;
                unsafeWindow.gmonkey.load('1.0', function(gm){ // This is the line that causes the major rich-text loading problems.
                    // For some reason, adding a callback function alleviates it.
                    // This is after prolonged isolating and testing. 
                });   
            } else {
                if(unsafeWindow.gmonkey.isLoaded('1.0')) {
                    pageWrapper._greasemonkeyLoaded(unsafeWindow.gmonkey.get('1.0'));
                    return; // Don't call setTimeout
                }
            }
        } catch(e){
            if( log ) log('ic error: ' + e, 'injected-code._loadGreasemonkey');
        }

        setTimeout(function() { pageWrapper._loadGreasemonkey(); }, 200);
    },

    /**
     * Takes the given gmonkey object and stores it before passing handling on
     * to the <code>_tryNotifyGreasemonkeyLoaded</code> function.
     * @function ?
     * @param {Gmonkey API} gmonkey - The Gmonkey API object
     */
    _greasemonkeyLoaded: function(gmonkey) {
        pageWrapper._gmonkey = gmonkey;
        pageWrapper._tryNotifyGreasemonkeyLoaded();
    },

    /**
     * Tries to register a view change callback on the gmonkey object, and if
     * successful, notifies the {@link glPageWrapper} object that the UI is ready.
     * @function ?
     */
    _tryNotifyGreasemonkeyLoaded: function() {
        try {
            pageWrapper._gmonkey.registerViewChangeCallback(function(view) { pageWrapper._viewChanged(view); });
        } catch(e) {
            // Gmail code not loaded yet :( - try again later
            setTimeout(function() { pageWrapper._tryNotifyGreasemonkeyLoaded(); }, 100);
            return;
        }

        // Is indeed loaded now
        pageWrapper.gmonkeyLoaded = true;
        this._dispatchEvent('ui_ready');
        
        // View changed is important for the view-dependant code to load correctly. It is not enough for code to manage loading itself within classes upon construction (which is a maintenance burden anyway); as the user does not always start in the ThreadList view - they could jump direct to the Conversation view. In short, we need to commence with a view_changed when we load.
        this._viewChanged();	// glPageWrapper.VIEW_CHANGED
    },

    /**
     * View change callback registered on the gmonkey object. Adds the class
     * "glViewRoot" to the active view element and notifies the
     * {@link glPageWrapper} object that the view has changed.
     * @function ?
     * @param {String} view - The active view type
     */
    _viewChanged: function(view) {
        // Change element with glViewRoot class
        var elems = document.getElementById('canvas_frame').contentDocument.getElementsByClassName('glViewRoot');
        for(var i = 0; i < elems.length; i++) {
            elems[i].className = elems[i].className.replace(/\s*glViewRoot\s*/, '');
        }
        var root = this._gmonkey.getActiveViewElement();
        if(root.className == "") {
            root.className = 'glViewRoot';
        } else {
            root.className += ' glViewRoot';
        }

        // Notify glPageWrapper
        this._dispatchEvent('view_changed');	// glPageWrapper.VIEW_CHANGED
    },
    //}

    /**
     * Dispatches an event on the injected element that the {@link glPageWrapper}
     * object can see, allowing information to be sent out of the page.
     * @function ?
     * @param {String} eventName - The event to dispatch
     */
    _dispatchEvent: function(eventName) {
        var elem = pageWrapper._getElem();
        elem.setAttribute('event', eventName);

        var evt = document.createEvent('Events');
        evt.initEvent('glGmailEventDispatched', true, false);
        elem.dispatchEvent(evt);

        elem.removeAttribute('event');
    },

    /**
     * "Called" by the {@link glPageWrapper} object, it runs the given function,
     * json encodes the result, and stores it in the attributes of the injected
     * function so that they can be extracted.
     * @function ?
     * @param {Event} evt - The glGmailRunFunction event
     */
    _runFunctionListener: function(evt) {
        var elem = pageWrapper._getElem();

        // Get function name and parameters from attributes
        var func = elem.getAttribute('func');
        var numParams = parseInt(elem.getAttribute('params'));
        var params = [];
        for(var i = 0; i < numParams; i++) {
            params.push(elem.getAttribute('param'+(i+1)));
        }

        // Run the given function
        var returnValue = pageWrapper[func].apply(null, params);

        // Convert return value to JSON and stick in element if defined
        pageWrapper._log('Return value of '+func+': '+returnValue);
        if(returnValue !== undefined) {
            try {
                elem.setAttribute('return', pageWrapper.JSON.stringify({'data':returnValue})); // Neeed object wrapper for native JSON
            } catch(e) {
                pageWrapper._log('Error in JSON encoding:'+"\n"+e);
            }
        }
    },

    /**
     * Returns the injected element, creating it if it doesn't exist.
     * @function {Element} ?
     * @return The injected communication element
     */
    _getElem: function() {
        if(!pageWrapper._elem) {
            var elem = document.createElementNS("http://www.w3.org/1999/xhtml","html:div");
            elem.setAttribute('id', '_glGmailInjected');
            elem.firebugIgnore = true;
            elem.setAttribute('style', 'display:none');
            document.documentElement.appendChild(elem);
            pageWrapper._elem = elem;
        }

        return pageWrapper._elem;
    },

    _log: function(data) {
        log(data, 'injected-code.js');
    }
};

// Get JSON encoding support - code from http://www.json.org/json2.js
if(window.JSON) {
    // Firefox 3.1 builtin
    pageWrapper.JSON = JSON;
} else {
    // No builtin JSON
    pageWrapper.JSON = {};
    var cx=/[\u0000\u00ad\u0600-\u0604\u070f\u17b4\u17b5\u200c-\u200f\u2028-\u202f\u2060-\u206f\ufeff\ufff0-\uffff]/g,escapable=/[\\\"\x00-\x1f\x7f-\x9f\u00ad\u0600-\u0604\u070f\u17b4\u17b5\u200c-\u200f\u2028-\u202f\u2060-\u206f\ufeff\ufff0-\uffff]/g,gap,indent,meta={'\b':'\\b','\t':'\\t','\n':'\\n','\f':'\\f','\r':'\\r','"':'\\"','\\':'\\\\'},rep;
    function quote(string){escapable.lastIndex=0;return escapable.test(string)?'"'+string.replace(escapable,function(a){var c=meta[a];return typeof c==='string'?c:'\\u'+('0000'+a.charCodeAt(0).toString(16)).slice(-4)})+'"':'"'+string+'"'}
    function str(key,holder){var i,k,v,length,mind=gap,partial,value=holder[key];if(typeof rep==='function'){value=rep.call(holder,key,value)}switch(typeof value){case'string':return quote(value);case'number':return isFinite(value)?String(value):'null';case'boolean':case'null':return String(value);case'object':if(!value){return'null'}gap+=indent;partial=[];if(typeof value.length == "number" && typeof value.splice != "undefined"){length=value.length;for(i=0;i<length;i+=1){partial[i]=str(i,value)||'null'}v=partial.length===0?'[]':gap?'[\n'+gap+partial.join(',\n'+gap)+'\n'+mind+']':'['+partial.join(',')+']';gap=mind;return v}if(rep&&typeof rep==='object'){length=rep.length;for(i=0;i<length;i+=1){k=rep[i];if(typeof k==='string'){v=str(k,value);if(v){partial.push(quote(k)+(gap?': ':':')+v)}}}}else{for(k in value){if(Object.hasOwnProperty.call(value,k)){v=str(k,value);if(v){partial.push(quote(k)+(gap?': ':':')+v)}}}}v=partial.length===0?'{}':gap?'{\n'+gap+partial.join(',\n'+gap)+'\n'+mind+'}':'{'+partial.join(',')+'}';gap=mind;return v}}
    if(typeof pageWrapper.JSON.stringify!=='function'){pageWrapper.JSON.stringify=function(value,replacer,space){var i;gap='';indent='';if(typeof space==='number'){for(i=0;i<space;i+=1){indent+=' '}}else if(typeof space==='string'){indent=space}rep=replacer;if(replacer&&typeof replacer!=='function'&&(typeof replacer!=='object'||typeof replacer.length!=='number')){throw new Error('JSON.stringify');}return str('',{'':value})}}
    if(typeof pageWrapper.JSON.parse!=='function'){pageWrapper.JSON.parse=function(text,reviver){var j;function walk(holder,key){var k,v,value=holder[key];if(value&&typeof value==='object'){for(k in value){if(Object.hasOwnProperty.call(value,k)){v=walk(value,k);if(v!==undefined){value[k]=v}else{delete value[k]}}}}return reviver.call(holder,key,value)}cx.lastIndex=0;if(cx.test(text)){text=text.replace(cx,function(a){return'\\u'+('0000'+a.charCodeAt(0).toString(16)).slice(-4)})}if(/^[\],:{}\s]*$/.test(text.replace(/\\(?:["\\\/bfnrt]|u[0-9a-fA-F]{4})/g,'@').replace(/"[^"\\\n\r]*"|true|false|null|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?/g,']').replace(/(?:^|:|,)(?:\s*\[)+/g,''))){j=eval('('+text+')');return typeof reviver==='function'?walk({'':j},''):j}throw new SyntaxError('JSON.parse');}};
}

pageWrapper.init();
})();