/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * A wrapper object around a UI widget in the gmail DOM. When connected, it
 * finds all important DOM elements and saves them in attributes.
 *
 * @class {abstract} ?
 */
var glUIWidget = glBase.extend({
    // Properties {
    /**
     * Whether or not it is connected. Useful for only running connect once or
     * re-connecting when it gets disconnected.
     * @property {Boolean} ?
     */
    _isConnected: null,
    
    /**
     * @property {glGmail} ?
     */
    _glGmail: null,
    
    //}

    /**
     * The constructor.
     * @constructor ?
     */
    constructor: function() {
        if(this.constructor === glUIWidget) throw new Error('Cannot be instantiated');
    },

    // Methods {
    /**
     * Connects the element to the given glGmail object's page. Called automatically
     * by {@link glGmail.getUIObject}.
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    connect: function(gmail) {
    	this._glGmail = gmail;
        this._isConnected = true;
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
    	this._glGmail = null;
        this._isConnected = false;
    },

    /**
     * Whether the UI object can be cached by {@link glGmail.getUIObject}
     * @function {Boolean} ?
     * @return Whether the UI can be cached
     */
    isCacheable: function() {
        return true;
    },

    /**
     * Returns whether or not connected
     * @function {Boolean} ?
     * @return Connection status
     */
    isConnected: function() {
        return this._isConnected;
    },
    //}
}, {
    // Static Methods {
    /**
     * Simulates the given mouse event on the given node
     * @function {static} ?
     * @param {Element} node - The node to simulate the event on
     * @param {String} eventType - The event type: mousedown, mouseup, ...
     */
    simulateMouseEvent: function(node, eventType) {
        var event = node.ownerDocument.createEvent("MouseEvents");
        event.initMouseEvent(
            eventType,
            true, // can bubble
            true, // cancellable
            node.ownerDocument.defaultView,
            1, // clicks
            50, 50, // screen coordinates
            50, 50, // client coordinates
            false, false, false, false, // control/alt/shift/meta
            0, // button,
            (eventType == 'mouseover' || eventType == 'mouseout') ? node : null
        );
        node.dispatchEvent(event);
    },
    
	addClassName: function(el, cn){
		if( glUIWidget.hasClassName(el,cn) ) return;
		if( el.className ) cn = el.className + " " + cn;
		el.className = cn;
	},
	
	removeClassName: function(el, cn){
		el.className = el.className.replace(cn, "");
		el.className = el.className.replace(/\s+$/, "");
	},
	
	hasClassName: function(el, cn){
		var re = new RegExp("(^|\\s)" + cn + "(\\s|$)");
		return re.test( el.className );
	},
	
	createGmailButton: function(name, v, options, doc){
		switch(name){
			case "gmail-header-button":
				var el = doc.createElement("DIV");
					el.className = "goog-imageless-button goog-inline-block";
					el.setAttribute("role", "button");
					el.setAttribute("tabindex", "0");
					el.setAttribute("onfocus", "javascript:this.className+=' goog-imageless-button-focused'");
					el.setAttribute("onblur", "javascript:this.className=this.className.replace(' goog-imageless-button-focused', '')");
					if( options && options.collapseLeft ) el.className += " " + "goog-imageless-button-collapse-left";
					if( options && options.collapseRight ) el.className += " " + "goog-imageless-button-collapse-right";
					el.innerHTML = "<div class='goog-inline-block goog-imageless-button-outer-box'><div class='goog-inline-block goog-imageless-button-inner-box'><div class='goog-imageless-button-pos'><div class='goog-imageless-button-top-shadow'> </div><div class='goog-imageless-button-content'>" + v + "</div></div></div></div>"
				return el;
		}
	},
	
	cumulativeOffset: function(element) {
		var valueT = 0, valueL = 0;
		do {
			valueT += element.offsetTop  || 0;
			valueL += element.offsetLeft || 0;
			element = element.offsetParent;
		} while (element);
		return [valueL, valueT];
	},
	
	getDimensions: function(element){
		var doc = element.ownerDocument;
		if( element.style.display!="none" ) return {width: element.offsetWidth, height: element.offsetHeight};
		return {width: 0, height: 0};
	},
	
	viewport: function(doc, useScrollOffset, incScrollBars){
		var win = doc.defaultView;
		var browserWidth = 0;
		var browserHeight = 0;
		if( incScrollBars ){
			browserWidth = win.innerWidth;
			browserHeight = win.innerHeight;
		}else{
			browserWidth = doc.documentElement.clientWidth;
			browserHeight = doc.documentElement.clientHeight;		
		}

		if( useScrollOffset ){
			browserWidth -= win.pageXOffset;
			browserHeight -= win.pageYOffset;
		}
		
		return [browserWidth, browserHeight];
	},
		
	
	getBorderWidth: function(el){
		// Return the width, in pixels, of the border
		var css = el.ownerDocument.defaultView.getComputedStyle(el,null);
		var borderWidth = el.style.borderWidth || (css? css.getPropertyValue('borderWidth'):null);
		if( !borderWidth ) borderWidth = el.style.border || (css? css.getPropertyValue('border'):null);
		var borderStyle = el.style.borderStyle || (css? css.getPropertyValue('borderStyle'):null);
		if( borderStyle && borderStyle.indexOf("none")==-1 && borderWidth && (borderWidth = borderWidth.match(/(\d)px/)) ){
			borderWidth = parseInt(borderWidth[1]);
		}else{
			borderWidth = 0;
		}
		return borderWidth;
	},
	
    
    //}
});