/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Listen to changes to threads (labels/stars) and optionally enforce rules (e.g. only one status label)
 *
 * @class gtdThreadRules @extends glBase
 */
var gtdThreadRules = glBase.extend({
    // Properties {
    /**
     * The gtdMainPagePlugin object for the page
     * @property {gtdMainPagePlugin} ?
     */
    _gtd: null,

	/** @property {Object[]} ? */
	_recentActions: null,

    //}

    /**
     * The constructor.
     * @constructor ?
     * @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
     */
    constructor: function(gtd) {
        this.base(gtd.gmail);
        this._gtd = gtd;
	this._recentActions = [];
		
    	this._gtd.gmail.actions.addEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);

    },

    // Methods {
	
	/**
	* Notification from action manager of an event.
	* If environmentGenerated, ensure we have already processed (code-request -> environment-generated-callback). If not, process it as a DOM event (as much be user generated).
	* Otherwise, process the event with whatever serverActionType originally used
	* @function {} ?
	*/
	_threadChanged: function(event){
			if( !event.thread ) return;
		
		
		
			if( event.environmentGenerated ){
				// Generated by environment. Check not already processed. If not, it must be DOM event, so process as such.
				var alreadyProcessed = false;
				for( var i = this._recentActions.length - 1; i>=0; i-- ){
					var recentEvent = this._recentActions[i].event;
					if( recentEvent.action==event.action && recentEvent.label==event.label ){	// TODO be ok for Star events too
						if( (recentEvent.threadObject && recentEvent.threadObject.hasId(event.thread)) || recentEvent.thread==event.thread ){	// recently processed
							alreadyProcessed = true;
							this._recentActions.splice(i,1);	// Remove all possible versions of this event
						}
					}
				}
				if( !alreadyProcessed ){
					this._processAction(event, glActionManager.USE_DOM);
				}
				
			}else{
				// Generated in code (by glActionManager.performAction)
				this._processAction(event, event.serverActionType )
				this._recentActions.push( {event: event, date: (new Date()).valueOf() } );
			}
			
			
			// Clear old recentActions
			var d = (new Date()).valueOf() - (1000*60*2);
			for( var i = this._recentActions.length-1; i>=0; i-- ){
				if( this._recentActions.date < d ) this._recentActions.splice(i,1);
			}

	},
	
	/**
	* Apply the rules to a given thread - based on what action was taken - and update with serverActionType
	* @function {} ?
	* @param {Object} event - The event from glActionManager
	* @param {String} serverActionType - The server action type to use to update the server
	*/
	_processAction: function(event, serverActionType){
		var statusData = this._gtd.labelsData.getStatusData(true);
		
		switch(event.action){
			case glActionManager.ADD_LABEL:
				if( statusData.reverseLookup[event.label] && event.label!=statusData.today ){	// Remove all other status labels
					var statuses = this._gtd.labelsData.getLabelsArray('statuses', false, true);
					for( var i = 0; i < statuses.length; i++ ){	
						if( statuses[i]!=event.label && statuses[i]!=statusData.today ){
							if( !event.threadObject || event.threadObject.hasLabel(statuses[i]) ){ // TODO Make this better at understanding where a thread (event.thread - optional) does not have a status label
								this._gtd.gmail.actions.performAction(glActionManager.REMOVE_LABEL, (event.threadObject || event.thread), serverActionType, statuses[i]);
							}
						}
					}
				}
				
				if( event.environmentGenerated && statusData.reverseLookup[event.label]=='finished' ){
					// Finish, generated by a user DOM click
					this._gtd.gmail.actions.performAction(glActionManager.MARK_AS_READ, event.thread, glActionManager.USE_DOM);
					this._gtd.gmail.actions.performAction(glActionManager.ARCHIVE, event.thread, glActionManager.USE_DOM);
				}
				
				break;
			case glActionManager.STAR:
				break;
			case glActionManager.UNSTAR:
				break;
		}
	},
	
    //}
});
