/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * The self-contained threads viewer for any given item in Gmail.
 *
 * @class ? @extends gtdUIObject
 */
var gtdExplorerThreadsViewer = gtdUIObject.extend({
    // Properties {
    
    
	/** @property {private gtdMainPagePlugin} ? The main gtd plugin for this page */
	_gtd: null,
	    	
    	/** @property {private Object} ? A collection of bundles. Format: {id:gtdExplorerBundle}*/
    	_bundles: null,
		
	/** @property {private Object} ? A collection of bundle ids for each type. Format: {type:String[]}*/
	_bundlesByType: null,
    	

    	/** @property {private Function} ? Optional filter for the item */
    	_itemFilter: null,
    	
    	/** @property {private gtdExplorerThreadsItem} ? The explorer item */
    	_threadsItem: null,
    	
    	/** @property {private gtdExplorerViewSwitch} ? The view switch */
    	_viewSwitch: null,

	/** @property {private String} ? The method for updating the server */
	_serverActionType: null,
    	
	/** @property {private glEventsContainer} ? Collection of manageable events on DOM elements */
	_eventsContainer: null,
    
	/**
	* The thread list container
	* @property {Element} ?
	*/
	el: null,
	
	/** @property {private Element} ?  */
	_elHeader: null,
	
	/** @property {private Element} ?  */
	_elBody: null,
	
	/** @property {private Element} ? */
	_elTabs: null,
	
	/** @property {private Element} ? */
	_elTabsList: null,
	
	/** @property {private Element} ?  */
	_elView: null,
	
	/** @property {private Element} ?  */
	_elWarning: null,
	
	
	
	
    //}


	// Methods {
	/**
	* Connects the element to the given gtdMainPagePlugin object's page. Called
	* automatically by {@link gtdMainPagePlugin.getUIObject}.
	* @function ?
	* @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
	*/
	connect: function(gtd) {
		glLogger.log('Connecting gtdExplorerThreadsViewer object to view', 'gtdExplorerThreadsViewer');
		
		this._gtd = gtd;
		this._eventsContainer = new glEventsContainer(this);
		this.base(gtd);
		
		// Create container
		var doc = gtd.gmail.canvasDocument;
		this.el = doc.createElement("DIV");
		this.el.id = "gtdi-dash";
		
		// Header
		this._elHeader = doc.createElement("DIV");
		this._elHeader.id = "gtdi-header";
		this.el.appendChild( this._elHeader );
		
		// Header / Items
		this._threadsItem = gtd.getUIObject("explorerThreadsItem");
		this._elHeader.appendChild( this._threadsItem.el );
		
		// Header / ViewSwitch
		this._viewSwitch = gtd.getUIObject("explorerViewSwitch");
		this._viewSwitch.addEventListener(gtdExplorerViewSwitch.CHANGE_VIEW, this._changeView, this);
		this._elHeader.appendChild( this._viewSwitch.el );
		
		// Body
		this._elBody = doc.createElement("DIV");
		this._elBody.id = "gtdi-body";
		this.el.appendChild( this._elBody );
			
		// Body / Tabs
		this._elTabs = doc.createElement("DIV");
		this._elTabs.id = "gtdi-menu";
		this._elBody.appendChild( this._elTabs );
		
		this._elTabsList = doc.createElement("UL");
		this._elTabs.appendChild( this._elTabsList );
		
		
		// Body / Thread
		this._elView = doc.createElement("DIV");
		this._elBody.appendChild( this._elView );
	
		this._bundles = {};
		this._bundlesByType = {};
		
		// Add event listeners
		this._eventsContainer.observe( this._elHeader, 'mousedown', this._clickHeader, false );
		this._eventsContainer.observe( this._elTabs, 'click', this._clickTabs, false );
	},



	/**
	* Called on page shutdown, it does cleanup
	* @function ?
	*/
	disconnect: function() {
		glLogger.log('Disconnecting gtdExplorerThreadsViewer from view', 'gtdExplorerThreadsViewer');
		
		if( this.el && this.el.parentNode ) this.el.parentNode.removeChild(this.el);
		this.el = null;
		
		// Wipe references to elements
		for( p in this ){
			if( this[p] && this[p].nodeName && this[p].nodeType ){
				this[p] = null;
			}
		}
		
		if( this._bundle ){
			if( this._bundle.getThreadListContainer && this._bundle.getThreadListContainer() ){
				this._bundle.getThreadListContainer().getThreadList().removeEventListener(gtdThreadList.LOAD_THREAD, this._loadThread, this);
			}
		}
		
		for( b in this._bundles ) this._bundles[b].unload();
		this._bundle = null;
		this._bundles = null;
		this._bundlesByType = null;
		
		this._itemFilter = null;
		
		
		this._eventsContainer.reset();
		this._eventsContainer = null;
		
		this._threadsItem.disconnect();
		this._threadsItem = null;
		
		this.base();
	},
	
	addBundle: function(bundle, types){
		this._bundles[bundle.id] = bundle;

		if( typeof types=='string' ) types = [types];
		for( var i = 0; i < types.length; i++ ){
			if( !this._bundlesByType[types[i]] ) this._bundlesByType[types[i]] = [];
			this._bundlesByType[types[i]].push(bundle.id);
		}
		
	},
	
	
	/**
	* Initiate the loading of this viewer.
	* Give tabs their initial counts and pass in an item filter for this item; rely on tabs to pass back a SELECT_TAB for the first item, which finishes the loading in ._tabChange
	* @function ?
	* @param {String} type - The type. Based on gtdExplorerThreadsViewer static properties
	* @param {String} item - The item. Depends on type (e.g. for gtdExplorerThreadsViewer.STATUSES it will be a status label. For gtdExplorerThreadsViewer.CONTACT it will be a name/email)
	*/
	load: function(type, item){
		
		if( this._elWarning ){
			this._elWarning.parentNode.removeChild( this._elWarning );
			this._elWarning = null;
		}
		
		if( !this._instantiated ){
			// Load bundles
			
			
			var all = new gtdExplorerBundleTlcAll(this._gtd);
			var today = new gtdExplorerBundleTlcToday(this._gtd);
			var recent = new gtdExplorerBundleTlcRecent(this._gtd);
			var tasksToDo = new gtdExplorerBundleTlcTasksToDo(this._gtd);
			var tasksWaitingOn = new gtdExplorerBundleTlcTasksWaitingOn(this._gtd);
			var tasksSomeDay = new gtdExplorerBundleTlcTasksSomeDay(this._gtd);
			var resources = new gtdExplorerBundleTlcResources(this._gtd);
			
			this.addBundle(all, [gtdExplorerThreadsViewer.STATUSES, gtdExplorerThreadsViewer.SEARCH]);
			this.addBundle(today, [gtdExplorerThreadsViewer.STATUSES, gtdExplorerThreadsViewer.PROJECTS, gtdExplorerThreadsViewer.CONTEXTS, gtdExplorerThreadsViewer.REFERENCES, gtdExplorerThreadsViewer.MISC, gtdExplorerThreadsViewer.SEARCH, gtdExplorerThreadsViewer.DASHBOARD, gtdExplorerThreadsViewer.CONTACT]);
			this.addBundle(tasksToDo, [gtdExplorerThreadsViewer.PROJECTS, gtdExplorerThreadsViewer.CONTEXTS, gtdExplorerThreadsViewer.REFERENCES, gtdExplorerThreadsViewer.MISC, gtdExplorerThreadsViewer.SEARCH, gtdExplorerThreadsViewer.DASHBOARD, gtdExplorerThreadsViewer.CONTACT]);
			this.addBundle(tasksWaitingOn, [gtdExplorerThreadsViewer.PROJECTS, gtdExplorerThreadsViewer.CONTEXTS, gtdExplorerThreadsViewer.REFERENCES, gtdExplorerThreadsViewer.MISC, gtdExplorerThreadsViewer.SEARCH, gtdExplorerThreadsViewer.DASHBOARD, gtdExplorerThreadsViewer.CONTACT]);
			this.addBundle(tasksSomeDay, [gtdExplorerThreadsViewer.PROJECTS, gtdExplorerThreadsViewer.CONTEXTS, gtdExplorerThreadsViewer.REFERENCES, gtdExplorerThreadsViewer.MISC, gtdExplorerThreadsViewer.SEARCH, gtdExplorerThreadsViewer.DASHBOARD, gtdExplorerThreadsViewer.CONTACT]);
			this.addBundle(resources, [gtdExplorerThreadsViewer.PROJECTS, gtdExplorerThreadsViewer.CONTEXTS, gtdExplorerThreadsViewer.REFERENCES, gtdExplorerThreadsViewer.MISC, gtdExplorerThreadsViewer.SEARCH, gtdExplorerThreadsViewer.CONTACT]);
			this.addBundle(recent, [gtdExplorerThreadsViewer.PROJECTS, gtdExplorerThreadsViewer.CONTEXTS, gtdExplorerThreadsViewer.REFERENCES, gtdExplorerThreadsViewer.MISC, gtdExplorerThreadsViewer.CONTACT]);
			
			this._instantiated = true;
		}
		
		// Clear the tabs:
		this._elTabsList.innerHTML = "";
		
		// Update the title
		this._threadsItem.setType(type, item);
		
		// De-activate existing bundles:
		if( this._type ){
			if( this._bundle ){
				if( this._bundle.getThreadListContainer && this._bundle.getThreadListContainer() ) this._bundle.getThreadListContainer().getThreadList().removeEventListener(gtdThreadList.LOAD_THREAD, this._loadThread, this);
			}
			for( var i = 0; i < this._bundlesByType[this._type].length; i++ ){
				var bundle = this._bundles[this._bundlesByType[this._type][i]];
				bundle.disconnect();
			}
			this._bundle = null;
		}
		
		
		
		// Load filters:
		this._itemFilter = null;
		switch( type ){
			case gtdExplorerThreadsViewer.STATUSES:
				this._itemFilter = function(thread){	// Test for presence of item
					return thread.hasLabel(item);
				};
				break;
			case gtdExplorerThreadsViewer.PROJECTS:
			case gtdExplorerThreadsViewer.CONTEXTS:
			case gtdExplorerThreadsViewer.REFERENCES:
			case gtdExplorerThreadsViewer.MISC:
				this._itemFilter = function(thread){	// Test for presence of item
					return thread.hasLabel(item);
				};
				break;
			case gtdExplorerThreadsViewer.CONTACT:
				this._itemFilter = function(thread){	// Test for presence of item TODO - requires thread messaging
					return thread.hasContact(item, item);
				};
				break;
		}
		
		
		// Connect the bundles:
		var firstBundle = null;
		for( var i = 0; i < this._bundlesByType[type].length; i++ ){
			var bundle = this._bundles[this._bundlesByType[type][i]];
			if( bundle.testCondition() ){
				bundle.connect(type, item, this._itemFilter);
				this._elTabsList.appendChild( bundle.getTabEl() );
				if( !firstBundle ) firstBundle = bundle;
			}
		}
		
		// Store variables:
		this._type = type || this._type;
		this._item = item || this._item;
		
		
		// Select first:
		if(firstBundle){
			this._loadBundle(firstBundle.id);
		}else{
			// No tabs loaded. Show message about configuring labels.
			this._elWarning = this._elView.ownerDocument.createElement("DIV");
			this._elWarning.className = "gtdi-no-tabs-warning";
			this._elWarning.innerHTML = glI18N.getString("Dashboard.NoTabsWarning.line1") + "<br/>" + glI18N.getString("Dashboard.NoTabsWarning.line2");
			this._elView.appendChild( this._elWarning );
		}
		
	},
	
	
	
	_clickTabs: function(event){
		try{
			var el = event.target;
			while( el.parentNode && !el.getAttribute("gtdi-tabid") ) el = el.parentNode;
			var id = el.getAttribute("gtdi-tabid");
			if( id ){
				glBase.stopEvent(event);
				this._loadBundle(id);
			}
		}catch(e){
			glLogger.warn(e, 'gtdExplorerThreadsViewer._clickTabs', this._gtd.gmail);
		}
	},
	
	_loadBundle: function(id){
		// Check different tab
		if( !this._bundle || id!=this._bundle.id ){ 
			// Disconnect old bundle
			if( this._bundle ){
				this._elView.innerHTML = "";
				this._bundle.selected(false);
				if( this._bundle.getThreadListContainer && this._bundle.getThreadListContainer() ) this._bundle.getThreadListContainer().getThreadList().removeEventListener(gtdThreadList.LOAD_THREAD, this._loadThread, this);
			}
			
			// Attach new bundle
			this._bundle = this._bundles[id];
			this._bundle.selected(true);
			this._bundle.setServerActionType(this._serverActionType);
			this._elView.appendChild( this._bundle.getViewEl() );
			
			// Check and sync all Bundles after loading one
			// Why? Because, in case of Search, it won't load thread container until first item selected. At that point, all other bundles that could benefit from that threadContainer will not know one sibling has it. So, they cannot update their counts. syncThreadList tells them to attempt to share; and thus hook into correct thread count. 
			for( var i = 0; i < this._bundlesByType[this._type].length; i++ ){
				var bundle = this._bundles[this._bundlesByType[this._type][i]];
				bundle.syncThreadContainer()
			}
			
			// Find the default view: 
			var view = gtdExplorerViewSwitch.PROJECT_VIEW;
			switch( this._type ){
				case gtdExplorerThreadsViewer.SEARCH:
					switch(id){
							case gtdExplorerBundle.ALL:
								view = gtdExplorerViewSwitch.LIST_VIEW;
								break;
					}
					break;
			}
			switch( id ){
				case gtdExplorerBundle.RECENT:
					view = gtdExplorerViewSwitch.LIST_VIEW;
					break;
			}
			this._viewSwitch.setView(view);
			this._bundle.getThreadListContainer().setView(view);	// TODO remove this expectation that it is a Tlc. Hide view objects if not Tlc. 
	
			this._bundle.getThreadListContainer().getThreadList().addEventListener(gtdThreadList.LOAD_THREAD, this._loadThread, this);
		}
	},
	
	/**
	* Notification that a view button was clicked. Pass the view to the threadlist
	* @function ?
	*/
	_changeView: function(event){
		if( this._bundle ) this._bundle.getThreadListContainer().setView(event.view); // TODO as above - remove expectation of Tlc
	},
	
	
	
	/**
	* Notification that a thread was clicked in the threadContainer
	* Pass it up
	* @function {} ?
	*/
	_loadThread: function(event){
		if( this.hasEventListener(gtdExplorerThreadsViewer.LOAD_THREAD)) this.dispatchEvent({name: gtdExplorerThreadsViewer.LOAD_THREAD, thread:event.thread});
	},
	
	_clickHeader: function(event){
		this.dispatchEvent({name: gtdExplorerThreadsViewer.CLICK_HEADER, mouseEvent: event});
	},
	
	
	
	/**
	* Get the title item
	* @function {gtdExplorerThreadsItem} ?
	*/
	getViewerItem: function(){
		return this._threadsItem;
	},
	
	/**
	* Set method for updating server. Ensure future created objects can use this info; and existing objects are updated.
	* @function ?
	* @param {String} type - The type. A static variable of glActionManager
	*/
	setServerActionType: function(type){
		this._serverActionType = type;
	},
	
	
	/**
	* Get the content currently in view
	* @function ?
	*/
	getCurrentThreadListContainer: function(){
		if( this._bundle && this._bundle.getThreadListContainer ) return this._bundle.getThreadListContainer();
	},
	
	/**
	* Whether the UI object can be cached by {@link gtdMainPagePlugin.getUIObject}
	* @function {Boolean} ?
	* @return Whether the UI can be cached
	*/
	isCacheable: function() {
		return false;
	},
	

}, {
	
	// Events {
	
	CLICK_HEADER: 'clickHeader',
	LOAD_THREAD: 'loadThread',
	
	
	//}


	DEFAULT_NUMBER_THREADS: 200,
	
	
	// Principle Types {
	
	STATUSES: 'status',
	PROJECTS: 'project',
	CONTEXTS: 'context',
	REFERENCES: 'reference',
	MISC: 'misc',
	SEARCH: 'search',
	DASHBOARD: 'dashboard',
	CONTACT: 'contact',
	
	// }
	
});
gtdExplorerThreadsViewer.implement(glEventDispatcher); // Add event dispatching support