/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Modify Labels box to break into GTD sections. 
 * Also store hierarchial label data. 
 *

 * @class ? @extends gtdUIObject
 */
var gtdLabelsBox = gtdUIObject.extend({
    // Properties {

    /** @property {private glShowHider[]} ? Array of ShowHiders in use  */
    _showHiders: null,

    /** @property {private glLabelsBox} ? The glLabelsBox  */
    _glLabelsBox: null,

    /** @property {private gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page  */
    _gtd: null,
    

    /** @property {Object} typeBuckets - Collection of LabelData (gtdLabelDOM) objects, grouped by type (e.g. Statuses). */
    typeBuckets: null,
    
    /** @property {Object} typeContainerEls - Collection of Table objects for each type */
    typeContainerEls: null,
    
    
    /** @property {private glEventsContainer} ? Collection of manageable events on DOM elements */
    _eventsContainer: null,
    
    /** @property {private glEventsContainer} ? Collection of manageable events on DOM elements */
    _eventsContainer2: null,
    
    //}
    

    // Methods {


    /**
    * @constructor ?
    * @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
    */
    connect: function(gtd){
        glLogger.log('Constructing gtdLabelsBox', 'gtdLabelsBox');
        this.base(gtd);
        this._gtd = gtd;
        this.typeBuckets = {};
        this.typeContainerEls = {};
        this._showHiders = [];
        this._glLabelsBox = gtd.gmail.getUIObject('labelsBox');
        this._glLabelsBox.addEventListener(glLabelsBox.LABELS_BOX_MODIFIED, this._arrangeLabelsBox, this);
                
        this._eventsContainer = new glEventsContainer(this);
        this._eventsContainer2 = new glEventsContainer(this);
        this._arrangeLabelsBox();
        glLogger.log('Constructed gtdLabelsBox', 'gtdLabelsBox');
    },
    
    /**
    * Clean up
    * @function ?
    */
    disconnect: function(){
        this._gtd = null;
        this._eventsContainer.reset();
        this._eventsContainer2.reset();
        for( var i = 0; i < this._showHiders.length; i++ ) this._showHiders[i].reset();
        this.base();
    },
    
    /**
    * Break the Labels box into GTD sections and extract hierarchial label data.
    * @function ?
    */
    _arrangeLabelsBox: function(){
    
    try{
        if( !this._glLabelsBox.getLabelsContainer() ) return;
        
        var doc = this._gtd.gmail.canvasDocument;
        var typeSettings = this._gtd.gmail.accountData.getPref('plugins.gtd.prefs.labels.types');
        
        glLogger.log('Start arrangeLabelsBox', 'gtdLabelsBox');
        var gtdArranged = this._glLabelsBox.getLabelsContainer().getAttribute("gtdArranged");
        if( gtdArranged ){	// Already processed or processing
            if( gtdArranged=="1" ){ // processed
                for( t in this.typeBuckets ){
                    for( l in this.typeBuckets[t].labelHash ){
                        var elLeaf = this.typeBuckets[t].getLeaf(l);
                        if( elLeaf.innerHTML.indexOf( typeSettings[t].prefix )==0 ){
                            elLeaf.innerHTML = elLeaf.innerHTML.replace(/.+?(\s(\(|\[)\d+(\)|\]))?$/, elLeaf.getAttribute("gtdi-leaf") + "$1");
                        }
                    }
                }
            }
            return;
        }
        this._glLabelsBox.getLabelsContainer().setAttribute("gtdArranged", "-1");	// Indicate processing
        
        
        glLogger.log('Process arrangeLabelsBox', 'gtdLabelsBox');
        
        this._eventsContainer2.reset();

                                        
        // Destroy old variables:
        var oldLabelsHash = this._gtd.labelsData? this._gtd.labelsData.getLabelsHash(null, true) : {};
        var labelsChange = {'id':(new Date()).valueOf(), 'changed':false};

        for( var i = 0; i < this._showHiders.length; i++ ){
            this._showHiders[i].reset();
        }
        this._showHiders.length = 0;
        
        if( this.typeContainerEls ){
            for( t in this.typeContainerEls ){
                delete this.typeContainerEls[t];
            }
            delete this.typeContainerEls;
        }
        
        if( this.typeBuckets ){
            for( t in this.typeBuckets ){
                this.typeBuckets[t].disconnect();
            }
            delete this.typeBuckets;
        }
        this.typeBuckets = {};



        // Grab the actual original table (contains each label as row)
        this.elTable = this._glLabelsBox.getLabelsTable();
        
        // Create New Structure for Labels Box:
        this.typeContainerEls = {};
        for( t in typeSettings ){
            this.typeContainerEls[t] = {'elTable':doc.createElement("TABLE"), 'elTBody':doc.createElement("TBODY")};
                glUIWidget.addClassName( this.typeContainerEls[t].elTable, "cf" );
                glUIWidget.addClassName( this.typeContainerEls[t].elTable, "pQ" );
                this.typeContainerEls[t].elTable.appendChild( this.typeContainerEls[t].elTBody );
                this.typeContainerEls[t].elTBody.className = "showhider-tbody";
            var sh = new glShowHider( this._gtd.gmail, "labelBox-" + t, {doc:doc, elParent:this._glLabelsBox.getLabelsContainer(), elBefore:this.elTable, elContent:this.typeContainerEls[t].elTable, noContentIndent:true, title:glI18N.getString("Labels.Types." + t, t)} );
                sh.elTitle.style.color = "black";
                sh.elTitle.style.fontWeight = "bold";
                sh.labelType = t;
                //sh.addEventListener( this._glLabelsBox.TOGGLE, this._toggleShowHider, this );
            this._showHiders.push( sh );
        }
        
        
        // Extract labels
        var elRows = this._glLabelsBox.getLabelItems();
        
        
        var fCreatePlaceHolderEl = function(labelname, leafname){
            var el = doc.createElement("TR");
            el.className = "WnjAod";
            var el2 = doc.createElement("TD");
            el2.innerHTML = "<div style='overflow: hidden; white-space: nowrap;'><div class='p5' gtdi-container='1'><span gtdi-leaf=''>"+leafname+"</span></div></div>";
            el2.firstChild.firstChild.firstChild.setAttribute("gtdi-leaf", leafname);
            var el3 = doc.createElement("TD");
            el3.className = "p7";
            el.appendChild( el2 );
            el.appendChild( el3 );
            return el;
        };
        for( var i = 0; i < elRows.snapshotLength; i++ ){
            
            var elLeaf = elRows.snapshotItem(i);	// A element
            var elRow = elLeaf.parentNode;	// Will be TR element
            while( elRow.parentNode && elRow.nodeName!="TR" ){
                    elRow = elRow.parentNode;
            }
            
            // Get the label name (- the number)
            var labelname = glBase.decodeEntities(elLeaf.innerHTML.replace(/\s\(\d+\)$/, ""), doc);
            
            
            // Find the type:
            var foundType = "";
            for( t in typeSettings ){
                if( gtdLabelsData.testPrefix(labelname, typeSettings[t].prefix) ){
                    foundType = t;
                    break;
                }
            }
            if( !foundType ) foundType = "misc";
            
            
            // Get the leaf name:
            var hierarchy = labelname.replace(typeSettings[t].prefix, "").split("/");
            var leafname = elLeaf.innerHTML.replace(/.+?(\s\(\d+\))?$/, hierarchy[hierarchy.length-1] + "$1");
            
            
            // Create a gtdLabelDOM container if needed; use the newly created TBODY to house it
            if( !this.typeBuckets[foundType] ){
                this.typeBuckets[foundType] = this._gtd.getUIObject('labelsTree');
                this.typeBuckets[foundType].init(foundType, this.typeContainerEls[t].elTBody);
            }
            
            this.typeBuckets[foundType].addLabel( elRow, elLeaf.parentNode, elLeaf, labelname, leafname, fCreatePlaceHolderEl );
            
            // Add Popout Icon
            if( !this.typeBuckets[foundType].isPlaceHolder(labelname) ){
                var elPopout = doc.createElement("DIV");
                elPopout.innerHTML = "<img src='chrome://gtdinbox/skin/gtd/filter-extend.png' class='gtdi-labeldom-popouticon'/>";
                elPopout.setAttribute("style", "position:absolute;width:9px;right:0px;top:0px;");	// TODO move to CSS
                elPopout.title = glI18N.getString("Labels.Popout.Title");
                
                elRow.getElementsByTagName("DIV")[0].style.position = "relative";
                elRow.getElementsByTagName("DIV")[0].firstChild.style.maxWidth = "124px";	// Second level DIV after TD. 136-9. 
                elRow.getElementsByTagName("DIV")[0].appendChild( elPopout );	// Add to first DIV after TD
            }
                
        
                
        }
        // Finalise label addition:
        for( t in this.typeBuckets ){
            this.typeBuckets[t].finishAddingLabels();
        }
        
        // Rearrange Statuses
        if( this.typeBuckets.statuses ){
            var defaultStatuses = [];
            for( d in typeSettings.statuses.core ){
                defaultStatuses.push( typeSettings.statuses.prefix + typeSettings.statuses.core[d] )
            }
            defaultStatuses.reverse();
            for( var i = 0; i < defaultStatuses.length; i++ ){
                if( this.typeBuckets.statuses.labelHash[defaultStatuses[i]] ){
                    this.typeContainerEls.statuses.elTBody.insertBefore( this.typeBuckets.statuses.labelHash[defaultStatuses[i]], this.typeContainerEls.statuses.elTBody.firstChild );
                }
            }		
        }
        
        
        // See if need to hide any empty boxes
        for( var i = 0; i < this._showHiders.length; i++ ){
            var sh = this._showHiders[i];
            if( this.typeContainerEls[sh.labelType].elTBody.innerHTML.length==0 ){	// Nothing doing
                sh.visible(false);
            }else{
                sh.visible(true);
            }
        }
        
        
        
        this._eventsContainer2.observe( this._glLabelsBox.getLabelsContainer(), "mousedown", this.labelsSectionMousedown, true);
        
        
        
        //this._gtdinboxPage._labelsBoxCount.refresh(true); TODO restore this
        
        
        if( oldLabelsHash ){
            for( t in this.typeBuckets ){
                for( l in this.typeBuckets[t].labelHash ){
                    if( oldLabelsHash[l] ){
                        delete oldLabelsHash[l];
                    }else{
                        labelsChange.added = true;
                    }
                }
            }
    
            if( glBase.hashLength(oldLabelsHash) > 0 ){	// Something deleted or renamed
                labelsChange.removed = true;
            }
        }
        
        if( labelsChange.added && this.hasEventListener(gtdLabelsBox.LABEL_ADDED)) this.dispatchEvent({name: gtdLabelsBox.LABEL_ADDED});
        if( labelsChange.removed && this.hasEventListener(gtdLabelsBox.LABEL_REMOVED)) this.dispatchEvent({name: gtdLabelsBox.LABEL_REMOVED});
                        
        this._glLabelsBox.getLabelsContainer().setAttribute("gtdArranged", "1");	// Indicate processed
        
        glLogger.log('Completed arrangeLabelsBox', 'gtdLabelsBox');
    }catch(e){
            glLogger.error(e, 'gtdLabelsBox.arrangeLabelsBox');
    }
                            
    },
    
    /**
    * Respond to a click event on the Labels box.
    * @function ?
    */
    labelsSectionMousedown: function(event, typeBucket){
            
        glLogger.log('Clicked labelsSectionMousedown', 'gtdLabelsBox');
        
        try{
            if( event && event.button!=0 ) return;
    
            var elClicked = event.target;

            if( elClicked.className=="gtdi-labeldom-popouticon" ){
                var elInspected = elClicked;
                while( elInspected.parentNode ){
                    elInspected = elInspected.parentNode;
                    if(( elInspected.className == 'WnjAod') || glUIWidget.hasClassName(elInspected, 'gtdi-labeldom-placeholder') || glUIWidget.hasClassName(elInspected, "gtdi-labeldom-label") ){
                        var labelName = elInspected.getAttribute("gtdLabelName");
                        var type = this._gtd.labelsData.getLabelType(labelName).toUpperCase();
        
                        // Get the popup
                        var explorerPopup = this._gtd.getUIObject("explorerPopup");

                        var threadsViewer = explorerPopup.getThreadsViewer(true);

                        // Get the x/y below the Dashboard button:
                        var xy = explorerPopup.getXYFromElement(elInspected, "right", true);
                        xy = explorerPopup.boundsCheckXY(xy[0], xy[1], null, true);
                        explorerPopup.active(true, xy[0], xy[1]);
                        explorerPopup.setServerActionType(glActionManager.USE_AJAX);
                        
                        // Load the Dashboard into the ThreadsViewer
                        threadsViewer.load( gtdExplorerThreadsViewer[type], labelName, this._gtd.statusThreadsContainer );
                
                        break;
                    }
                }
            }
        }catch(e){
                glLogger.warn(e, 'gtdLabelsBox.labelsSectionMousedown', this._gtd.gmail);
        }
    },
    
    
    /**
    * Whether the UI object can be cached by {@link gtdMainPagePlugin.getUIObject}
    * @function {Boolean} ?
    * @return Whether the UI can be cached
    */
    isCacheable: function() {
            return false;
    },
    
    
},{
    // Static Properties {

    /**
     * Dispatched when new label detected (will not fire on first run)
     * @event ?
     */
    LABEL_ADDED: 'label_added',
    
    /**
     * Dispatched when label detected removed (will not fire on first run)
     * @event ?
     */
    LABEL_REMOVED: 'label_removed',
    //}
    
});

gtdLabelsBox.implement(glEventDispatcher); // Add event dispatching support