/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * A list of threads, possibly partitioned by project.
 *
 * @class ? @extends gtdUIObject
 */
var gtdThreadList = gtdUIObject.extend({
    // Properties {
    
	/**
	* The thread list container
	* @property {Element} ?
	*/
	el: null,

	/** @property {private Object} ? Elements for the loading boxes. Format: {loading:{el:, elCurrent:, elTotal:}, partialComplete:{el:, elCurrent:, elTotal:}} */
	_loadbox: null,

	/**
	* The thread list container, minus the loading box
	* @property {Element} ?
	*/	
	_elInner: null,
    
	/** @property {private gtdMainPagePlugin} ? The main gtd plugin for this page */
	_gtd: null,
	
	/** @property {private glEventsContainer} ? Collection of manageable events on DOM elements */
	_eventsContainer: null,
	
	/** @property {private Object} ? A collection of functions that accepts a thread and returns true if it is to be rendered. Format: {<key>:Function} */
	_filters: null,
	
	/** @property {private Object} ? A hash of thread arrays {'general':{el:, tlThreads:gtdTLThread[]}, 'p:/gtdinbox':{el:, tlThreads:gtdTLThread[]}}. (Nb 'general' would only be used if not partitioning into projects) */
	_uiContainers: null,
	
	/** @property {private Object} ? A hash of all tlThread objects; keyed on threadId. Used for partial reloads. */
	_tlThreadHash: null,
	
	/** @property {private gtdThreadContainer} ? The thread container */
	_threadContainer: null,
	
	/** @property {private boolean} ? True if thread container is passed in. Not to be 'unload'ed and disables fresh searches. */
	_sharedThreadContainer: null,
	
	/** @property {private boolean} ? If true, split threads into their projects */
	_partitionByProject: null,
	
	/** @property {private Boolean} ? True if currently visible */
	_active: null,
    //}


	// Methods {
	/**
	* Connects the element to the given gtdMainPagePlugin object's page. Called
	* automatically by {@link gtdMainPagePlugin.getUIObject}.
	* @function ?
	* @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
	*/
	connect: function(gtd) {
		glLogger.log('Connecting gtdThreadList object to view', 'gtdThreadList');
		
		this._gtd = gtd;
		this.base(gtd);
		this._eventsContainer = new glEventsContainer(this);
		this._partitionByProject = true;
		this._uiContainers = {};
		this._tlThreadHash = {};
		this._filters = {};
		
		var doc = gtd.gmail.canvasDocument;
		this.el = doc.createElement("DIV");
		this.el.id = "gtdi-threadlists";
		
		// Create the loading boxes
		this._loadbox = {loading:{}, partialComplete:{}};
		
		this._loadbox.loading.el = doc.createElement("DIV");
		this._loadbox.loading.el.style.display = "none";
		this._loadbox.loading.el.className = "gtdi-loadingbox";
		this._loadbox.loading.el.innerHTML = "<img src='chrome://gtdinbox/skin/indicator_small.gif'/><div>" + glI18N.getString("ThreadList.Loadbox.loading").replace("%1", "<span></span>").replace("%2", "<span></span>") + "</div>";
		this._loadbox.loading.elCurrent = this._loadbox.loading.el.getElementsByTagName("SPAN")[0];
		this._loadbox.loading.elTotal = this._loadbox.loading.el.getElementsByTagName("SPAN")[1];
		this.el.appendChild( this._loadbox.loading.el );
		
		this._loadbox.partialComplete.el = doc.createElement("DIV");
		this._loadbox.partialComplete.el.className = "gtdi-loadingbox";
		this._loadbox.partialComplete.el.style.display = "none";
		this._loadbox.partialComplete.el.innerHTML = glI18N.getString("ThreadList.Loadbox.partialComplete").replace("%1", "<span></span>").replace("%2", "<span></span>");
		this._loadbox.partialComplete.elCurrent = this._loadbox.partialComplete.el.getElementsByTagName("SPAN")[0];
		this._loadbox.partialComplete.elTotal = this._loadbox.partialComplete.el.getElementsByTagName("SPAN")[1];
		this._loadbox.partialComplete.elDownload = doc.createElement("A");
		this._loadbox.partialComplete.elDownload.className = "gtdi-downloadall";
		this._loadbox.partialComplete.elDownload.href = "#";
		this._loadbox.partialComplete.elDownload.innerHTML = glI18N.getString("ThreadList.Loadbox.partialComplete.all");
		this._loadbox.partialComplete.el.appendChild( this._loadbox.partialComplete.elDownload );
		this.el.appendChild( this._loadbox.partialComplete.el );
		
		// Create inner, where the thread ui goes
		this._elInner = doc.createElement("DIV");
		this.el.appendChild( this._elInner );
				
		// Add listeners:
		this._eventsContainer.observe(this.el, "click", this._click, false);		
	},

	/**
	* Called on page shutdown, it does cleanup
	* @function ?
	*/
	disconnect: function() {
		glLogger.log('Disconnecting gtdThreadList from view', 'gtdThreadList');
		
		if( this.el && this.el.parentNode ) this.el.parentNode.removeChild(this.el);
		this.el = null;
		this._elInner = null;
		this._loading = null;
		this._eventsContainer.reset();
		
		if( this._threadContainer ){
			this._threadContainer.removeEventListener(gtdThreadContainer.NEW_SEARCH, this._newSearch, this);
			this._threadContainer.removeEventListener(gtdThreadContainer.LOADED, this._loadComplete, this);
			this._threadContainer.removeEventListener(gtdThreadContainer.LOADING_UPDATE, this._loadingUpdate, this);
			this._threadContainer.removeEventListener(gtdThreadContainer.CHANGED, this._threadsChanged, this);
			if( !this._sharedThreadContainer ) this._threadContainer.unload();	// Do not unload shared ones; may be used elsewhere.
			this._threadContainer = null;
		}
		
		if( this._uiContainers ){
			for( c in this._uiContainers ){
				for( var i = 0; i < this._uiContainers[c].tlThreads.length; i++ ){
					this._uiContainers[c].tlThreads[i].disconnect();
				}
			}
		}
		this._uiContainers = null;
		this._tlThreadHash = null;
		this._eventsContainer = null;
		this._filters = null;
		
		this.base();
	},
	
	/**
	* Decide intensity with which it uses resources. If in view, set active.
	* @function {Boolean} ?
	* @param {Boolean} v - If true, set active
	* @return If v is omitted, return active status
	*/
	active: function(v){
		if( typeof v=="undefined" ) return this._active;
		
		this._active = v;
		if( v ){
			this._threadContainer.addEventListener(gtdThreadContainer.LOADING_UPDATE, this._loadingUpdate, this);
		}else{
			this._threadContainer.removeEventListener(gtdThreadContainer.LOADING_UPDATE, this._loadingUpdate, this);
			this._threadContainer.cancelSearch();
		}
	},
	
	/**
	* Load a new thread container and render it
	* Wipes out any filters in place.
	* @function ?
	* @param {glSearch} search - The search to populate the container
	* @param {gtdThreadContainer} threadContainer - Optional. A shared thread container to use instead of creating own. 
	*/
	load: function(search, threadContainer, debugId){ // TODO - remove debugId
		
		// Clean up old thread container
		if( this._threadContainer ){
			this._threadContainer.removeEventListener(gtdThreadContainer.NEW_SEARCH, this._newSearch, this);
			this._threadContainer.removeEventListener(gtdThreadContainer.LOADED, this._loadComplete, this);
			this._threadContainer.removeEventListener(gtdThreadContainer.LOADING_UPDATE, this._loadingUpdate, this);
			this._threadContainer.removeEventListener(gtdThreadContainer.CHANGED, this._threadsChanged, this);
			if( !this._sharedThreadContainer ) this._threadContainer.unload();	// Do not unload shared ones; may be used elsewhere.
			this._threadContainer = null;
		}
		
		for( f in this._filters ) delete this._filters[f];
		
		// Establish the new thread container
		if( threadContainer ){
			// Attach to shared one
			this._threadContainer = threadContainer;
			this._threadContainer.addEventListener(gtdThreadContainer.NEW_SEARCH, this._newSearch, this);
			this._threadContainer.addEventListener(gtdThreadContainer.LOADED, this._loadComplete, this);
			if( this._active ) this._threadContainer.addEventListener(gtdThreadContainer.LOADING_UPDATE, this._loadingUpdate, this);
			var v = this._threadContainer.addEventListener(gtdThreadContainer.CHANGED, this._threadsChanged, this);
			this._sharedThreadContainer = true;
			this._reload();
		}else{
			// Create a new one
			this._threadContainer = gtdThreadContainer.createWithSearch(this._gtd, search);
			this._threadContainer.addEventListener(gtdThreadContainer.NEW_SEARCH, this._newSearch, this);
			this._threadContainer.addEventListener(gtdThreadContainer.LOADED, this._loadComplete, this);
			if( this._active ) this._threadContainer.addEventListener(gtdThreadContainer.LOADING_UPDATE, this._loadingUpdate, this);
			var v = this._threadContainer.addEventListener(gtdThreadContainer.CHANGED, this._threadsChanged, this);
			this._threadContainer.load();
			this._sharedThreadContainer = false;
		}
		
		this.active(true);
		
		this._debugId = debugId;
	},
	
	/**
	* Called when the gtdThreadContainer commences a new search
	* @function ?
	*/
	_newSearch: function(event){
				
		// Update the loading box
		this._loadbox.partialComplete.el.style.display = "none";
		var upto = this._threadContainer.threads.length+25;
		var total = this._threadContainer.getNumberThreads(gtdThreadContainer.COUNT_TOTAL_DESIRED);
		if( this._threadContainer.getNumberThreads(gtdThreadContainer.COUNT_TOTAL_CALCULATED) < total ) total = this._threadContainer.getNumberThreads(gtdThreadContainer.COUNT_TOTAL_CALCULATED);
		if( upto > total ) upto = total;
		this._loadbox.loading.elCurrent.innerHTML = this._threadContainer.threads.length + "-" + upto;
		this._loadbox.loading.elTotal.innerHTML = total;
		this._loadbox.loading.el.style.display = "";
	
		// Complete reload:
		this._reload();
		if( this.hasEventListener(event.name)) this.dispatchEvent(event); // propogate
	},
	
	/**
	* Called when the gtdThreadContainer has new threads during a search. Precipitates a refresh.
	* @function ?
	*/
	_loadingUpdate: function(event){
		
		// Update the loading box
		this._loadbox.partialComplete.el.style.display = "none";		
		var upto = this._threadContainer.threads.length+25;
		var total = this._threadContainer.getNumberThreads(gtdThreadContainer.COUNT_TOTAL_DESIRED);
		if( this._threadContainer.getNumberThreads(gtdThreadContainer.COUNT_TOTAL_CALCULATED) < total ) total = this._threadContainer.getNumberThreads(gtdThreadContainer.COUNT_TOTAL_CALCULATED);
		if( upto > total ) upto = total;
		this._loadbox.loading.elCurrent.innerHTML = this._threadContainer.threads.length + "-" + upto;
		this._loadbox.loading.elTotal.innerHTML = total;
		this._loadbox.loading.el.style.display = "";
		
		// TODO Change this to read, Downloading 25-50 of 200 (3000 total). Downloading 25-50 of 200 (200 total).
		
		// Partial reload:	
		this._reload(true);	// TODO Can we simply add the new elements? So as not to destroy the UI elements if people are working with them.
		if( this.hasEventListener(event.name)) this.dispatchEvent(event); // propogate
	},
	
	/**
	* Called when the gtdThreadContainer is loaded with a new search. Precipitates a refresh.
	* @function ?
	*/
	_loadComplete: function(event){
		
		// Update the loading box
		this._loadbox.loading.el.style.display = "none";
		if( this._threadContainer.isComplete() ){		
			this._loadbox.partialComplete.el.style.display = "none";
		}else{
			this._loadbox.partialComplete.elCurrent.innerHTML = this._threadContainer.threads.length;
			this._loadbox.partialComplete.elTotal.innerHTML = this._threadContainer.getNumberThreads(gtdThreadContainer.COUNT_TOTAL_CALCULATED);
			this._loadbox.partialComplete.el.style.display = "";
		}
	
	
		// Partial reload:
		this._reload(true);	// TODO Can we simply add the new elements? So as not to destroy the UI elements if people are working with them.
		if( this.hasEventListener(event.name)) this.dispatchEvent(event); // propogate
	},
	

	
	/**
	* Called when a thread is updated. Find the updated threads and request they refresh themselves.
	* @function ?
	* @param {Object} event - {name:, addedThreads:glThread[], updatedThreads:glThread[], removedThreads:glThread[]}
	*/
	_threadsChanged: function(event){
		
			if( event.addedThreads.length>0 ){
				// Full refresh - unknown structural impact
				// TODO doesn't actually date sort correctly. Probably a thread-order problem (and thus the responsibility of gtdThreadContainer._threadChanged) rather than our own date-sort issue.
				this._reload(true);
			}
			
			// Convert to hashes for faster/cleaner lookups (Nb cannot store threadlist thread's in a hash because a thread can appear in multiple containers - not 1-1)
			var updatedHash = {};
			for( var i = 0; i < event.updatedThreads.length; i++ ) updatedHash[event.updatedThreads[i].threadId] = event.updatedThreads[i];
			var removedHash = {};
			for( var i = 0; i < event.removedThreads.length; i++ ) removedHash[event.removedThreads[i].threadId] = event.removedThreads[i];
			
			
			// Test each thread to see if it requires refreshing or removing:
			for( c in this._uiContainers ){
				for( var i = this._uiContainers[c].tlThreads.length - 1; i>=0; i-- ){
					var tlThread = this._uiContainers[c].tlThreads[i];
					if( removedHash[tlThread.thread.threadId] ){		// Thread was removed. Clear it out.
						this._uiContainers[c].tlThreads.splice(i,1);
						this._tlThreadHash[tlThread.thread.threadId] = null;
						this._uiContainers[c].el.style.display = (this._getVisibleThreadCount( c )==0)? "none" : "";
						if( this._uiContainers[c].tlThreads.length==0 ){	// Remove container
							this._uiContainers[c].el.parentNode.removeChild( this._uiContainers[c].el );
							for( var j = 0; j < this._uiContainers[c].tlThreads.length; j++ ) this._uiContainers[c].tlThreads[j].disconnect();
							delete this._uiContainers[c];
							break;
						}
						tlThread.disconnect();
					}else if( updatedHash[tlThread.thread.threadId] ){	// Thread was updated
						tlThread.refresh();
					}
				}
			}
			if( this.hasEventListener(event.name)) this.dispatchEvent(event); // propogate
		
	},
	
	/**
	* Destroy existing thread list UI elements and rebuild for every visible element. Factor in filter.
	* @function ?
	* @param {Boolean} partial - If true, do not delete existing UI elements; just add new threads not yet listed
	*/
	_reload: function(partial){
		// Clear out anything previous:
				
		if( !partial ){
			for( c in this._uiContainers ){
				for( var i = 0; i < this._uiContainers[c].tlThreads.length; i++ ) this._uiContainers[c].tlThreads[i].disconnect();
			}
			this._uiContainers = {};
			this._tlThreadHash = {};
			this._elInner.innerHTML = "";
		}
		
		
		// Render threads in gtdThreadContainer
		var doc = this._gtd.gmail.canvasDocument;
		if( this._partitionByProject ){
			// Split threads into projects, then render each group
			var typeSettings = this._gtd.gmail.accountData.getPref('plugins.gtd.prefs.labels.types');
			var prefix = typeSettings.projects.prefix;
			var projectHash = {};	// Format: projectHash[<labelname>] = glThread[];
			for( var i = 0; i < this._threadContainer.threads.length; i++ ){	// Process each thread
				var foundProject = false;
				for( var j = 0; j < this._threadContainer.threads[i].labels.length; j++ ){	// Scan each thread label for projects
					var labelname = this._threadContainer.threads[i].labels[j];
					if( gtdLabelsData.testPrefix( labelname, prefix ) ){	// It's a project label
						if( !projectHash[labelname] ) projectHash[labelname] = [];
						projectHash[labelname].push( this._threadContainer.threads[i] );
						foundProject = true;
					}
				}
				if( !foundProject ){	// Associate with Misc
					if( !projectHash['general'] ) projectHash['general'] = [];
					projectHash['general'].push( this._threadContainer.threads[i] );
				}
			}
			// Now render each project as a uiContainer; and populate with threads
			var containerNames = [];
			for( p in projectHash ){
				if( !this._uiContainers[p] ){
					this._uiContainers[p] = {el:doc.createElement("DIV"), elList:doc.createElement("UL"), tlThreads:[]};
					this._uiContainers[p].elHeader = doc.createElement("H2");
					if( p=="general" ) this._uiContainers[p].elHeader.innerHTML = glI18N.getString("ThreadList.Headers.general");
					else this._uiContainers[p].elHeader.innerHTML = p;
					this._uiContainers[p].elList.className = "gtdi-threadlist";
					this._uiContainers[p].el.appendChild( this._uiContainers[p].elHeader );
					this._uiContainers[p].el.appendChild( this._uiContainers[p].elList );
					//this._elInner.appendChild( this._uiContainers[p].el );
				}
				
				for( var i = 0; i < projectHash[p].length; i++ ){
					var tlThread = this._tlThreadHash[ projectHash[p][i].threadId ];
					if( !tlThread  ){
						tlThread = this._gtd.getUIObject("tLThread");
						tlThread.load( projectHash[p][i] );	// Load the thread
						this._uiContainers[p].elList.appendChild( tlThread.el );
						this._uiContainers[p].tlThreads.push( tlThread );
						this._tlThreadHash[ projectHash[p][i].threadId ] = tlThread;
					}
					
					// Filter if neccessary. Only hide so can be shown again if filter changes
					this._testFilters( tlThread );
				}
				// If container no longer has visible threads, hide it
				containerNames.push(p);
			}
			// Sort Containers by Name:
			containerNames.sort().reverse();
			if( !this._elInner.firstChild ) this._elInner.innerHTML = "<div style='height:1px'></div>";
			var onlyGeneral = true;
			for( var i = 0; i < containerNames.length; i++ ){
				var p = containerNames[i];
				if( p=="general" ){
					this._elInner.appendChild( this._uiContainers[p].el );
				}else{
					this._elInner.insertBefore( this._uiContainers[p].el, this._elInner.firstChild );
				}
				if( this._getVisibleThreadCount(p)==0 ){
					this._uiContainers[p].el.style.display = 'none';
					if(p=='general') onlyGeneral = false;
				}else{
					this._uiContainers[p].el.style.display = '';
					if(p!='general') onlyGeneral = false;
				}
			}
			if( this._uiContainers['general'] ) this._uiContainers['general'].elHeader.style.display = onlyGeneral? 'none' : '';	// Do not show 'Misc' if the only one
		}else{
			// Create a single container
			
			if( !this._uiContainers['general'] ){
				this._uiContainers['general'] = {el:doc.createElement("DIV"), elList:doc.createElement("UL"), tlThreads:[]};
				this._uiContainers['general'].elList.className = "gtdi-threadlist";
				this._uiContainers['general'].elHeader = doc.createElement("DIV");
				this._uiContainers['general'].el.appendChild( this._uiContainers['general'].elList );
				this._elInner.appendChild( this._uiContainers['general'].el );
			}
			// Render a simple list
			for( var i = 0; i < this._threadContainer.threads.length; i++ ){	// Process each thread
				var tlThread = this._tlThreadHash[ this._threadContainer.threads[i].threadId ];
				if( !tlThread ){
					tlThread = this._gtd.getUIObject("tLThread");
					tlThread.load( this._threadContainer.threads[i] );	// Load the thread
					this._uiContainers['general'].elList.appendChild( tlThread.el );
					this._uiContainers['general'].tlThreads.push( tlThread );
					this._tlThreadHash[ this._threadContainer.threads[i].threadId ] = tlThread;
				}
				// Filter if neccessary. Only hide so can be shown again if filter changes
				this._testFilters(tlThread);
			}	
			this._uiContainers['general'].el.style.display = (this._getVisibleThreadCount( 'general' )==0)? "none" : "";
		}
		
		this._testShowNoThreads();
 	},
	
	/**
	* If conditions right (no threads, finished searching), show the No Threads message.
	* @function {} ?
	*/
	_testShowNoThreads: function(){
		var elNoThreads = this._elInner.ownerDocument.evaluate( ".//*[@class='gtdi-no-threads']", this._elInner, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue
		if( elNoThreads ) elNoThreads.parentNode.removeChild( elNoThreads );
		if( !this._threadContainer.isSearching() && this.getFilterCount()==0 ){
			var elNoThreads = this._elInner.ownerDocument.createElement("DIV");
			elNoThreads.className = "gtdi-no-threads";
			elNoThreads.innerHTML = glI18N.getString('ThreadList.NoThreads');
			this._elInner.appendChild( elNoThreads );
		}
	},
	
	/**
	* Get number of tlThreads that are visible in any given container (they may be hidden by a filter)
	* @function ?
	* @param {string} uiContainerName - The desired container name
	* @return {int} - The number of threads that are visible in this container
	*/
	_getVisibleThreadCount: function(uiContainerName){
		var c = 0; 
		for( var i = 0; i < this._uiContainers[uiContainerName].tlThreads.length; i++ ){
			if( this._uiContainers[uiContainerName].tlThreads[i].el.style.display!="none" ) c++;
		}
		return c;
	},
	
	/**
	* Set how the threads are rendered. List or Project.
	* @function ?
	* @param {String} id - List or Project, using static properties of gtdExplorerViewSwitch
	*/
	setView: function(id){
		var changed = false;
		if( id==gtdExplorerViewSwitch.PROJECT_VIEW ){
			if( !this._partitionByProject ) changed = true;
			this._partitionByProject = true;
		}else{
			if( this._partitionByProject ) changed = true;
			this._partitionByProject = false;
		}
		if( changed ) this._reload();
	},
	
	/**
	* Set the filter function for threads; and ensure thread visibility is set.
	* @function ?
	* @param {String} key - A unique ID for the filter function
	* @param {Function} filter - Function that accepts a thread and returns true if it should be retained
	*/
	setFilter: function(key, filter){
		if( filter ){
			this._filters[key] = filter;
		}else{
			if( this._filters[key] ) delete this._filters[key];
		}
		
		// Process each gtdTLThread and show/hide appropriately
		var onlyGeneral = true;
		for( c in this._uiContainers ){
			for( var i = 0; i < this._uiContainers[c].tlThreads.length; i++ ){
				this._testFilters(this._uiContainers[c].tlThreads[i]);
			}
			if(this._getVisibleThreadCount( c )==0){
				this._uiContainers[c].el.style.display = 'none';
				if(c=='general') onlyGeneral = false;
			}else{
				this._uiContainers[c].el.style.display = '';
				if(c!='general') onlyGeneral = false;
			}
		}
		if( this._uiContainers['general'] ) this._uiContainers['general'].elHeader.style.display = onlyGeneral? 'none' : '';
		this._testShowNoThreads();
	},
	
	/**
	* Get a filter for a specific key
	* @function {Function} ?
	* @param {String} key - A unique ID for the filter function
	* @return The filter function
	*/
	getFilter: function(key){
		return this._filters[key];
	},
	
	/**
	* See if a filter applies to a thread. If thread should be retained, ensure shown and return true. Otherwise hide it and mark it (to exclude from typed-search filtering).
	* @function ?
	* @param {gtdTLThread} tlThread - The thread to test
	* @return {boolean} - True if thread passes all filters
	*/
	_testFilters: function(tlThread){
		for( f in this._filters ){
			if( !this._filters[f](tlThread.thread) ){
				tlThread.el.style.display = "none";
				tlThread.el.setAttribute("gtd-filtered-out", "1");
				return false;
			}
		}
		tlThread.el.setAttribute("gtd-filtered-out", "");
		tlThread.el.style.display = "";
		return true;
	},
	
	/**
	* Get count of how many threads pass the filter, including all other active filters
	* Used by gtdThreadListerContainer to get counts for buttons.
	* @function ?
	* @param {Function} filter - The filter function. Return true if a thread should be retained.
	* @param {String} excludeKey - Exclude one of the environment filters. E.g. gtdThreadListContainer would exclude 'filter_button' when testing a potential 'filter_button' filter.
	* @return {int} - The number of matching threads
	*/
	getFilterCount: function(filter, excludeKey){		
		var c = 0; 
		for( var i = 0; i < this._threadContainer.threads.length; i++ ){
			var thread = this._threadContainer.threads[i];
			
			if( !filter || filter(thread) ){
				var match = true;
				for( f in this._filters ){
					if( f!=excludeKey && !this._filters[f](thread) ){
						match = false;
						break;
					}
				}
				if( match ) c++;
			}
		}
		return c;
	},
	
	/**
	* Get an array of threads that conform to all current filters
	* @function {glThread[]} ?
	*/
	getFilteredThreads: function(){
		var threads = [];
		for( var i = 0; i < this._threadContainer.threads.length; i++ ){
			var thread = this._threadContainer.threads[i];
			var match = true;
			for( f in this._filters ){
				if( !this._filters[f](thread) ){
					match = false;
					break;
				}
			}
			if( match ) threads.push(thread);
		}
		return threads;
	},
	
	/**
	* Hide any threads that do not contain, somewhere, the search text.
	* For speed, does not consider any gtdTLThreads that have already been filtered by this._filters
	* @function ?
	* @param {String} searchText - the text to find in threads. For threads that have it, leave them showing
	*/
	search: function(searchText){
		if( this._uiContainers ){
			for( c in this._uiContainers ){
				for( var i = 0; i < this._uiContainers[c].tlThreads.length; i++ ){
					var tlThread = this._uiContainers[c].tlThreads[i];
					if( !tlThread.getAttribute("gtd-filtered-out") ){
						// TODO test tlThread.thread and set display of tlThread.el
					}
				}
			}
		}
	},
	
	/**  
	* Get array of glThread objects that have been checked in the UI
	* @function ?
	* @return {glThread[]} - Array of threads that have been ticked in checkbox
	*/
	getCheckedThreads: function(){
		var checkedThreads = {};	// {<threadId>:glThread} - use hash to avoid thread duplication
		if( this._uiContainers ){
			for( c in this._uiContainers ){
				for( var i = 0; i < this._uiContainers[c].tlThreads.length; i++ ){
					var tlThread = this._uiContainers[c].tlThreads[i];
					if( tlThread.isChecked() ) checkedThreads[tlThread.thread.threadId] = tlThread.thread;
				}
			}
		}
		return glBase.hashToArray(checkedThreads);
	},
	
	/** 
	* Respond to a click anywhere in the thread list
	* Handle label Xing, star-ing, and clicking links to a thread.
	* @function ?
	*/
	_click: function(event){
		try{
			var el = event.target;
			
			
			if( el.className=="gtdi-downloadall" ){	// Download All loading link
				this._threadContainer.search();	// download everything in this query
			}else{
				
				
				// Extract the thread id:
				var elThread = el;
				while( elThread.parentNode && !elThread.getAttribute("gtdi-thread-id") ) elThread = elThread.parentNode;
				var threadid = elThread.getAttribute("gtdi-thread-id");
			
				// Test thread interaction:
				if( threadid ){
					var thread = this._threadContainer.getThreadById(threadid);
					if( el.className=="gtdi-labelremove" ){		// Label-X
						var labelname = el.parentNode.getAttribute("gtdi-labelname");
						this._gtd.gmail.actions.performAction( glActionManager.REMOVE_LABEL, thread, null, labelname );	// Notification of success will be detected the gtdThreadContainer, which will notify us to refresh that thread.
					}else if( el.parentNode.className=="gtdi-star" ){ 	// Star
						if( el.src=="chrome://gtdinbox/skin/gtd/i/gtdi-star-off.png" ){	// Turn it on
							this._gtd.gmail.actions.performAction( glActionManager.STAR, thread );
						}else{								// Turn it off
							this._gtd.gmail.actions.performAction( glActionManager.UNSTAR, thread );				
						}
					}else if( el.nodeName=="INPUT" ){	// Check box
						if( el.checked ){
							glUIWidget.addClassName( elThread, "gtdi-active" );
						}else{
							glUIWidget.removeClassName( elThread, "gtdi-active" );
						}
						if( this.hasEventListener(gtdThreadList.SELECT_THREAD)) this.dispatchEvent({name: gtdThreadList.SELECT_THREAD, thread:thread, checked:el.checked});
					}else{	// Thread
						// Broadcast
						if( this.hasEventListener(gtdThreadList.LOAD_THREAD)) this.dispatchEvent({name: gtdThreadList.LOAD_THREAD, thread:thread});
					}
				}
			}
		}catch(e){
			glLogger.warn(e, "gtdThreadList._click", this._gtd.gmail);
		}
	},
	
	/**
	* Get the data-orientated thread container that supports this view
	* @function {gtdThreadContainer} ?
	*/
	getThreadContainer: function(){
		return this._threadContainer;
	},
	
	
	/**
	* Whether the UI object can be cached by {@link gtdMainPagePlugin.getUIObject}
	* @function {Boolean} ?
	* @return Whether the UI can be cached
	*/
	isCacheable: function() {
		return false;
	},
	

    //}
},{
    // Static Properties {
    /**
     * Dispatched when a thread is clicked
     * @event ?
     */
    LOAD_THREAD: 'load_thread',
    
    
    /**
     * Dispatched when a checkbox is clicked
     * @event ?
     */
    SELECT_THREAD: 'clickCheckbox',
        
});
gtdThreadList.implement(glEventDispatcher); // Add event dispatching support





















/**
 * A line item in the thread list. Represents a thread.
 *
 * @class ? @extends gtdUIObject
 */
var gtdTLThread = gtdUIObject.extend({
    // Properties {
    
    
	/** @property {private gtdMainPagePlugin} ? The main gtd plugin for this page */
	_gtd: null,
    
	/**
	* The thread list container
	* @property {Element} ?
	*/
	el: null,
    
	/** @property {private Element} ? */
	_elCheckbox: null,
	
	/** @property {private Element} ? */
	_elStar: null,
	
	/** @property {private Element} ? */
	_elSubject: null,
	
	/** @property {private Element} ? */
	_elPeople: null,
	
	/** @property {private Element} ? */
	_elAttachment: null,
	
	/** @property {private Element} ? */
	_elDate: null,
	
	/** @property {private Element} ? */
	_elLabels: null,
		
	/** @property {private glThread} ? */
	thread: null,
    //}


	// Methods {
	/**
	* Connects the element to the given gtdMainPagePlugin object's page. Called
	* automatically by {@link gtdMainPagePlugin.getUIObject}.
	* @function ?
	* @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
	*/
	connect: function(gtd) {
		glLogger.log('Connecting gtdTLThread object to view', 'gtdTLThread');
		
		this._gtd = gtd;
		this.base(gtd);
		
		// Create container
		var doc = gtd.gmail.canvasDocument;
		this.el = doc.createElement("LI");

		// Create details
		var elDetails = doc.createElement("UL");
		elDetails.className = "gtdi-threaddetails";
		this.el.appendChild( elDetails );
		
			// Checkbox
			var elLi = doc.createElement("LI");
			elLi.className = "gtdi-checked";
		
			this._elCheckbox = doc.createElement("INPUT");
			this._elCheckbox.setAttribute("type", "checkbox");
			
			elLi.appendChild( this._elCheckbox );
			elDetails.appendChild( elLi );
			
			// Star
			var elLi = doc.createElement("LI");
			elLi.className = "gtdi-star";
		
			this._elStar = doc.createElement("IMG");
			
			elLi.appendChild( this._elStar );
			elDetails.appendChild( elLi );
			
			// Subject
			this._elSubject = doc.createElement("LI");
			this._elSubject.className = "gtdi-threadsubject";
	
			elDetails.appendChild( this._elSubject );
						
			// People
			this._elPeople = doc.createElement("LI");
			this._elPeople.className = "gtdi-people";
	
			elDetails.appendChild( this._elPeople );
			
			// Attachment
			var elLi = doc.createElement("LI");
			elLi.className = "gtdi-attachment";
		
			this._elAttachment = doc.createElement("IMG");
			this._elAttachment.src = "chrome://gtdinbox/skin/gtd/i/gtdi-attachment.png";
			this._elAttachment.title = glI18N.getString("TLThread.Attachment.title");
			this._elAttachment.style.display = "none";
			
			elLi.appendChild( this._elAttachment );
			elDetails.appendChild( elLi );
			
			// Date
			this._elDate = doc.createElement("LI");
			this._elDate.className = "gtdi-date";
	
			elDetails.appendChild( this._elDate );
			
		
		
		// Create labels container
		this._elLabels = doc.createElement("UL");
		this._elLabels.className = "gtdi-labels";
		this.el.appendChild( this._elLabels );
		
	},

	/**
	* Associate a thread with this UI element and update with the latest thread data
	* @function ?
	* @param {glThread} thread - The thread to render
	*/
	load: function(thread){
		this.el.setAttribute("gtdi-thread-id", thread.threadId);
		this.thread = thread;
		this.refresh();
	},
	
	/**
	* Rebuild the data elements from the thread. Preserve checkbox status.
	* @function ?
	*/
	refresh: function(){
		// Basic data:
		this._elSubject.title = glBase.decodeEntities(this.thread.subjectSnippet);
		this._elSubject.innerHTML = this._elSubject.title;
		this._elDate.innerHTML = this.thread.dateSnippet;	// TODO get proper date from thread messages and format as per gtd.properties
		this._elPeople.innerHTML = glBase.stripTags(this.thread.contactsSnippet);
		this.el.title = glBase.decodeEntities(glBase.stripTags(this.thread.bodySnippet));
		
		// Star status:
		if( this.thread.isStarred ){
			this._elStar.src = "chrome://gtdinbox/skin/gtd/i/gtdi-star-on.png";
			this._elStar.title = glI18N.getString("TLThread.star.on");
		}else{
			this._elStar.src = "chrome://gtdinbox/skin/gtd/i/gtdi-star-off.png";
			this._elStar.title = glI18N.getString("TLThread.star.off");
		}
		
		// Attachment
		if( this.thread.attachmentsCSV ){
			this._elAttachment.style.display = "";
			this._elAttachment.title = this.thread.attachmentsCSV;
		}else{
			this._elAttachment.style.display = "none";
		}
		
		
		// Add labels:
		var doc = this._gtd.gmail.canvasDocument;
		this._elLabels.innerHTML = "";
		var statusEls = [];
		var projectEls = [];
		var otherEls = [];
		var typeSettings = this._gtd.gmail.accountData.getPref('plugins.gtd.prefs.labels.types');
		for( var i = 0; i < this.thread.labels.length; i++ ){
			var labelname = this.thread.labels[i];
			if( labelname.indexOf("^")!=0 ){	// Avoid internal
				
				var elLi = doc.createElement("LI");
				var elLabel = doc.createElement("DIV");
				elLabel.className = "gtdi-label";
				elLabel.setAttribute("gtdi-labelname", labelname);
				elLabel.innerHTML = "<span class='gtdi-labelname'>" + labelname + "</span><span class='gtdi-labelremove'>x</span>";
				elLi.appendChild( elLabel );
				
				if( gtdLabelsData.testPrefix( labelname, typeSettings.statuses.prefix ) ){
					statusEls.push(elLi);
				}else if( gtdLabelsData.testPrefix( labelname, typeSettings.projects.prefix ) ){
					projectEls.push(elLi);
				}else{
					otherEls.push(elLi);
				}
			}
		}
		for( var i = 0; i < statusEls.length; i++ ) this._elLabels.appendChild( statusEls[i] );
		for( var i = 0; i < projectEls.length; i++ ) this._elLabels.appendChild( projectEls[i] );
		for( var i = 0; i < otherEls.length; i++ ) this._elLabels.appendChild( otherEls[i] );
	},

	/**
	* Find out if this thread has been ticked/selected
	* @function ?
	* @return {boolean} - True if ticked/selected
	*/
	isChecked: function(){
		return this._elCheckbox.checked;
	},


	/**
	* Whether the UI object can be cached by {@link gtdMainPagePlugin.getUIObject}
	* @function {Boolean} ?
	* @return Whether the UI can be cached
	*/
	isCacheable: function() {
		return false;
	},
	

	/**
	* Called on page shutdown, it does cleanup
	* @function ?
	*/
	disconnect: function() {
		glLogger.log('Disconnecting gtdTLThread from view', 'gtdTLThread');
		
		if( this.el && this.el.parentNode ) this.el.parentNode.removeChild(this.el);
		this.el = null;
		
		// Wipe references to elements
		for( p in this ){
			if( this[p] && this[p].nodeName && this[p].nodeType ){
				this[p] = null;
			}
		}
		
		this.thread = null;
		
		this.base();
	},

});
// TODO create a static factory function to create new TLThread items; and on disconnect, put them back on a stack so they can be reused. Otherwise we'll be creating thousands of elements in any Gmail session.
