/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Adds command buttons to the thread list
 *
 * @class ? @extends glUIView
 */
var gtdThreadListCommands = gtdUIObject.extend({
    // Properties {
	/** @property {private gtdMainPagePlugin} ? The main gtd plugin for this page */
	_gtd: null,
	
	/** @property {private glEventsContainer} ? Collection of manageable events on DOM elements */
	_eventsContainer: null,
		
	/** @property {Element} ? The thread list container */
	elTL: null,
	
	/** @property {Element[]} ? Array of Thread List tables parents*/
	tableContainerEls: null,
	
	/** @property {private glMultiToOne} ? Reduce rapid function calls into a single one per Xms */
    	_multiToOne: null,
    	
	/** @property {private Boolean[]} ? The rows checked before we kick in */
	_lastCheckedRows: null,
	
	/** @property {private glThread} */
	_currentThread: null,
    	
    //}

    // Methods {
    /**
     * Connects the element to the given glGmail object's page. Called automatically
     * by {@link glGmail.getUIObject}.
     * @function ?
     * @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
     */
    connect: function(gtd) {
        glLogger.log('Connecting gtdThreadListCommands object to view', 'gtdThreadListCommands');

        this.base(gtd);
        
        this._gtd = gtd;
        this.elTL = gtd.gmail.page.getViewRoot();
		this.tableContainerEls = [];
	this._eventsContainer = new glEventsContainer(this);
	this._multiToOne = new glMultiToOne(500);
	
	
	this._eventsContainer.observe( this.elTL, "mousedown", this._click, true);
	this._eventsContainer.observe( this.elTL, "contextmenu", this._click, true);
	
	this._refresh();
    },
	
	/**
	* Listen for changes to the ThreadList; and if changed, do refresh
	* @function ?
	* @param {boolean} v - If true, listen to changes. If not, ignore changes.
	*/
	_listenDOM: function(v){
		// Remove any listeners regardless
		if( this.tableContainerEls ){
			for( var i = 0; i < this.tableContainerEls.length; i++ ) this._eventsContainer.stopObserving(this.tableContainerEls[i], "DOMNodeInserted", this._domNodeInserted, false);
		}
		
		// Refresh the list of Tables
	        var elTLs = this._gtd.gmail.getCurrentViewUIObject().getThreadListTables();	
			this.tableContainerEls.length = 0;
			for( var i = 0; i < elTLs.length; i++ ) this.tableContainerEls[i] = elTLs[i].parentNode.parentNode;
	        
	        if( v ){
	        	// Listen to the tables
		        for( var i = 0; i < this.tableContainerEls.length; i++ ){
					this._eventsContainer.observe(this.tableContainerEls[i], "DOMNodeInserted", this._domNodeInserted, false);
				}
			}
	},
	
	/**
	* Reduce event-calls for changes to dom-node-inserted; turning any event calls within a 400ms into a single refresh call
	* @function ?
	*/
	_domNodeInserted: function(event){
		try{
			this._listenDOM(false);
			this._multiToOne.process("threadListDomChange", glBase.bind(this, this._refresh));
		}catch(e){
			glLogger.warn(e, 'gtdThreadListCommands._domNodeInserted');
		}
	},
	
	
	/**
	* Ensure we have cmd-icons added to each row in the thread list
	* @function ?
	*/
	_refresh: function(){
					
						
		try{
			this._listenDOM(false);
			
			var doc = this._gtd.gmail.canvasDocument;
			
			var elImg = doc.createElement("IMG");
			elImg.className = "gtd-row-button";
			elImg.style.position = "absolute";
			elImg.style.right = "12px";
			elImg.src = "chrome://gtdinbox/skin/gtd/gtd_lineitem.png";
			elImg.setAttribute("onmouseover", "this.src='chrome://gtdinbox/skin/gtd/gtd_lineitem-hover.png'");
			elImg.setAttribute("onmouseout", "this.src='chrome://gtdinbox/skin/gtd/gtd_lineitem.png'");
			elImg.title = glI18N.getString("ThreadListUI.GTDBox.Tooltip");
			
			var elImgBin = doc.createElement("IMG");
			elImgBin.className = "gtd-row-bin-button";
			elImgBin.style.position = "absolute";
			elImgBin.style.right = "0px";
			elImgBin.src = "chrome://gtdinbox/skin/gtd/bin_closed.png";
			elImgBin.setAttribute("onmouseover", "this.src='chrome://gtdinbox/skin/gtd/bin_closed-hover.png'");
			elImgBin.setAttribute("onmouseout", "this.src='chrome://gtdinbox/skin/gtd/bin_closed.png'");
			elImgBin.title = glI18N.getString("ThreadListUI.GTDBin.Tooltip");
			
			var previewEnabled = this._gtd.gmail.accountData.getPref('plugins.gtd.prefs.components.thread_list.preview_button.enabled');
			var trashEnabled = this._gtd.gmail.accountData.getPref('plugins.gtd.prefs.components.thread_list.trash_button.enabled');
			
			var result = this._gtd.gmail.getCurrentViewUIObject().getLastTDInRows();
			if( result.snapshotLength>0 ){
				if( (result.snapshotItem(0).innerHTML.indexOf("gtd")==-1 || result.snapshotItem(result.snapshotLength-1).innerHTML.indexOf("gtd")==-1) ){
					for( var i = 0; i < result.snapshotLength; i++ ){
						if( previewEnabled ){
						    var elImgC = elImg.cloneNode(true);
						    result.snapshotItem(i).appendChild( elImgC );
						}
						if( trashEnabled ){
						    var elImgBinC = elImgBin.cloneNode(true);
						    result.snapshotItem(i).appendChild( elImgBinC );
						}
					}
				}
			}
			
		}catch(e){
			glLogger.log('Error in gtdThreadListCommands.refresh -> ' + e.toString(), 'gtdThreadListCommands');
		}
		
		this._listenDOM(true);
		
		
	        glLogger.log('refreshed', 'gtdThreadListCommands');
	},
	
	
	/**
	* If clicked cmd icon, process the click
	* @function ?
	*/
	_click: function(event){
		try{
			// Retrieve what was clicked, and verify cmd icon
			var elCmd = event.target;
			if( !(elCmd.className=="gtd-row-button" || elCmd.className=="gtd-row-bin-button") ){
				// Did not click a button
				return;
			}
			
			glBase.stopEvent(event);
			
			var doc = this._gtd.gmail.canvasDocument;
			
			// Retrieve the gmail threadListView functionality
			var gmailThreadListView = this._gtd.gmail.getCurrentViewUIObject();
			
			// Uncheck all checkboxes
			this._lastCheckedRows = gmailThreadListView.checkedRows();
			gmailThreadListView.uncheckRows();
			
			// Retrieve associated checkbox
			var el = elCmd;
			while( el.parentNode && el.nodeName!="TR" ) el = el.parentNode;
			el = el.getElementsByTagName("INPUT")[0];
			if( el.checked==false ) glUIWidget.simulateMouseEvent(el, "click");	// Check it
			
			// Choose action
			if( elCmd.className=="gtd-row-button" ){
				if( event.button==0 && !event.ctrlKey ){
					if( event.shiftKey ){	//
					}else{			// Preview
						var thread = gmailThreadListView.getThread( gmailThreadListView.getOffset(elCmd) );
						this._currentThread = thread;
						
						// Get the popup
						var explorerPopup = this._gtd.getUIObject("explorerPopup");
						
						// Set the popup to use ThreadViewer and grab a reference to it
						var threadViewer = explorerPopup.getThreadViewer(true, null, null, false);
						threadViewer.load(null, thread);

						// Get the x/y below the Dashboard button:
						var xy = explorerPopup.getXYFromElement(elCmd, "bottomLeft", true);
						xy = explorerPopup.boundsCheckXY(xy[0], xy[1], null, true);
						explorerPopup.active(true, xy[0], xy[1]);
						explorerPopup.setServerActionType(glActionManager.USE_DOM);
						
						
						// Listen for thread changes
				    	this._gtd.gmail.actions.addEventListener(glActionManager.ARCHIVE.eventName, this._threadArchived, this);
				
						// Need to uncheck when closed
						explorerPopup.addEventListener(gtdExplorerPopup.HIDE, this._closeThreadPopup, this);
						
					}
				}else if( event.button==2 || event.ctrlKey ){	// Action label + archive
					var thread = gmailThreadListView.getThread( gmailThreadListView.getOffset(elCmd) );
					// add label
					var status = this._gtd.labelsData.getStatusData();
					if( this._gtd.labelsData.hasLabel(status.action) ) this._gtd.gmail.actions.performAction(glActionManager.ADD_LABEL, thread, glActionManager.USE_DOM, status.action )
					this._gtd.gmail.actions.performAction(glActionManager.MARK_AS_READ, thread, glActionManager.USE_DOM);
					this._gtd.gmail.actions.performAction(glActionManager.ARCHIVE, thread, glActionManager.USE_DOM);
				}
			}else if( elCmd.className=="gtd-row-bin-button" ){
				if( event.button==0 && !event.ctrlKey ){
					if( event.shiftKey ){	// Spam
						/*
						var elSpam = gmailThreadListView.actionsHeader.getButton(glActionsHeader.REPORT_SPAM_BUTTON);
						glUIWidget.simulateMouseEvent(elSpam, 'mousedown');
						glUIWidget.simulateMouseEvent(elSpam, 'mouseup');
						*/
						this._gtd.gmail.actions.performAction(glActionManager.REPORT_SPAM, thread, glActionManager.USE_DOM);
					}else{			// Finish label + archive
						var thread = gmailThreadListView.getThread( gmailThreadListView.getOffset(elCmd) );
						// add label
						var status = this._gtd.labelsData.getStatusData();
						if( this._gtd.labelsData.hasLabel(status.finished) ) this._gtd.gmail.actions.performAction(glActionManager.ADD_LABEL, thread, glActionManager.USE_DOM, status.finished )
						this._gtd.gmail.actions.performAction(glActionManager.MARK_AS_READ, thread, glActionManager.USE_DOM);
						this._gtd.gmail.actions.performAction(glActionManager.ARCHIVE, thread, glActionManager.USE_DOM);
					}
				}else if( event.button==2 || event.ctrlKey ){	// Delete
					this._gtd.gmail.actions.performAction(glActionManager.DELETE, thread, glActionManager.USE_DOM);
				}
			}
		}catch(e){
			glLogger.warn(e, 'gtdThreadListCommands._click', this._gtd.gmail);
		}
	},
	
	
	/**
	* Notification that a thread was archived, from glActionManager
	* @function {} ?
	*/
	_threadArchived: function(evt){
			if( this._currentThread && this._currentThread.hasId(evt.item) && evt.environmentGenerated ){
					var explorerPopup = this._gtd.getUIObject("explorerPopup");
					explorerPopup.active(false);
			}
	},
	
	/**
	* Notification that the thread popup has closed.
	* Remove listeners and uncheck
	* @function ?
	*/
	_closeThreadPopup: function(){
		
		var explorerPopup = this._gtd.getUIObject("explorerPopup");
		explorerPopup.removeEventListener(gtdExplorerPopup.HIDE, this._closeThreadPopup, this);
		
		// Restore checked rows:
		var gmailThreadListView = this._gtd.gmail.getCurrentViewUIObject();
		gmailThreadListView.checkedRows(this._lastCheckedRows);
		
    	this._gtd.gmail.actions.removeEventListener(glActionManager.ARCHIVE.eventName, this._threadArchived, this);
		this._currentThread = null;
	},

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        glLogger.log('Disconnecting gtdThreadListCommands from view', 'gtdThreadListCommands');
        
        this.elTL = null;
        this.tableContainerEls = null;
        this._eventsContainer.reset();
        this._multiToOne.reset();
        this.base();
    },
    
    /**
     * Whether the UI object can be cached by {@link gtdMainPagePlugin.getUIObject}
     * @function {Boolean} ?
     * @return Whether the UI can be cached
     */
    isCacheable: function() {
        return false;
    },
    

    //}
});
