/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * ***** END LICENSE BLOCK ***** */



/**
 * Wraps around a thread list to provide filters and actions over multiple threads.
 *
 * @class ? @extends gtdUIObject
 */
var gtdThreadListContainer = gtdUIObject.extend({
    // Properties {
    
    
	/** @property {private gtdMainPagePlugin} ? The main gtd plugin for this page */
	_gtd: null,
	
	/** @property {private glEventsContainer} ? Collection of manageable events on DOM elements */
	_eventsContainer: null,
    	   	
    
	/**
	* The thread list container
	* @property {Element} ?
	*/
	el: null,
	
	/** @property {private Element} ? The filter button menu */
	_elFiltersBar: null,
	
	/** @property {private Element} ? The thread actions menu(s) at the base */
	_elActionsBar: null,
	
	/** @property {private gtdThreadList} ? The UI for a collection of threads */
	_threadList: null,
	
	/** @property {private gtdLabelsMenu} ? The UI for managing a group of threads */
	_labelsMenu: null,
	
	/** @property {private Object} ? Filter elements for items in _elFiltersBar. Format: {<key>:Element} */
	_filterEls: null,
	
	/** @property {private Object} ? Filters in use. Format: {<key>:Function} */
	_filters: null,
	
	
    //}


	// Methods {
	/**
	* Connects the element to the given gtdMainPagePlugin object's page. Called
	* automatically by {@link gtdMainPagePlugin.getUIObject}.
	* @function ?
	* @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
	*/
	connect: function(gtd) {
		glLogger.log('Connecting gtdThreadListContainer object to view', 'gtdThreadListContainer');
		
		this._gtd = gtd;
		this._eventsContainer = new glEventsContainer(this);
		this.base(gtd);
		
		// Create container
		var doc = gtd.gmail.canvasDocument;
		this.el = doc.createElement("DIV");
		this.el.id = "gtdi-threads";
		
		// Create filter menu container:
		var elFilterMenu = doc.createElement("DIV");
		elFilterMenu.id = "gtdi-filter-menu";
		this.el.appendChild( elFilterMenu );
		
		this._elFiltersBar = doc.createElement("UL");
		elFilterMenu.appendChild( this._elFiltersBar );
		
		// Create Thread List
		this._threadList = this._gtd.getUIObject("threadList");
		this.el.appendChild( this._threadList.el );
		
		// Create actions bar:
		var elActions = doc.createElement("DIV");
		elActions.id = "gtdi-threadtoolbar";
		this.el.appendChild( elActions );
		
		this._elActionsBar = doc.createElement("DIV");
		this._elActionsBar.className = "gtdi-btns";
		elActions.appendChild( this._elActionsBar );
		
		this._labelsMenu = gtd.getUIObject("labelsMenu");
		this._elActionsBar.appendChild( this._labelsMenu.el );
		
		// Setup filters:
		this._createFilters();
		this._gtd.labelsData.addEventListener(gtdLabelsData.LABEL_CHANGED, this._createFilters, this);
		
		// Listen for thread events
		this._threadList.addEventListener(gtdThreadContainer.LOADED, this._threadsChanged, this);
		this._threadList.addEventListener(gtdThreadContainer.CHANGED, this._threadsChanged, this);
		this._threadList.addEventListener(gtdThreadList.LOAD_THREAD, this._selectThread, this);
		this._threadList.addEventListener(gtdThreadList.SELECT_THREAD, this._checkedThreads, this);
		
		// Listen for click events
		this._eventsContainer.observe( this._elFiltersBar, "click", this._clickFiltersBar, false );
	},



	/**
	* Called on page shutdown, it does cleanup
	* @function ?
	*/
	disconnect: function() {
            glLogger.log('Disconnecting gtdThreadListContainer from view', 'gtdThreadListContainer');
            
            if( this.el && this.el.parentNode ) this.el.parentNode.removeChild(this.el);
            this.el = null;
            
            // Wipe references to elements
            for( p in this ){
                if( this[p] && this[p].nodeName && this[p].nodeType ){
                    this[p] = null;
                }
            }
            
            
            this._gtd.labelsData.removeEventListener(gtdLabelsData.LABEL_CHANGED, this._createFilters, this);
            this._threadList.removeEventListener(gtdThreadContainer.LOADED, this._threadsChanged, this);
            this._threadList.removeEventListener(gtdThreadContainer.CHANGED, this._threadsChanged, this);
            this._threadList.removeEventListener(gtdThreadList.LOAD_THREAD, this._selectThread, this);
            this._threadList.removeEventListener(gtdThreadList.SELECT_THREAD, this._checkedThreads, this);
            
            this._threadList.disconnect();
            this._threadList = null;
            
            this._labelsMenu.disconnect();
            this._labelsMenu = null;
            
            this._eventsContainer.reset();
            this._eventsContainer = null;
            
            this._filterEls = null;
            this._filters = null;
            
            this.base();
	},


	/**
	* Load a new thread list and render it
	* @function ?
	* @param {glSearch} search - The search to populate the container
	* @param {gtdThreadContainer} threadContainer - Optional. A shared thread container to use instead of creating own. 
	*/
	load: function(search, threadContainer, debugId){	// TODO debugId
		this._threadList.load(search, threadContainer, debugId);
		this._refreshCounts();
	},
	
	
	/** 
	* Create new filter functions on every call
	* On first call, create filter button elements
	* @function ?
	*/
	_createFilters: function(){
		// Update filter functions:
		if( this._filters ) delete this._filters;
		var statusData = this._gtd.labelsData.getStatusData();
		var tasksToDo = [statusData.next_action, statusData.action, statusData.some_day];
		var references = this._gtd.labelsData.getLabelsArray("references", false, true);
		var labelsData = this._gtd.labelsData;
		this._filters = {
			'all': null,
			'next_action': function(thread){	// TODO Need conditions for filter tabs
					return thread.hasLabel( statusData.next_action );
			},
			'action': function(thread){
					return thread.hasLabel( statusData.action );
			},
			/*
			'some_day': function(thread){
					return thread.hasLabel( statusData.some_day );
			},
			*/
			'waiting_on': function(thread){
					return thread.hasLabel( statusData.waiting_on );
			},
			'personal': function(thread){
					return thread.isPersonal();
			},
			'unread': function(thread){
					if( !thread.isRead ) return true;
					return false;
			},
			/*
			'resources': function(thread){
					if( thread.attachmentsCSV ) return true;
					for( var i = 0; i < references.length; i++ ) if( thread.hasLabel( references[i] ) ) return true;
			},
			*/
		};
		
		this._conditions = {
			'next_action': function(){
				if( labelsData.hasLabel(statusData.next_action) ) return true;
			},
		};
		
		// Create elements (if first call)
		if( !this._filterEls ) this._filterEls = {};
		if( glBase.hashLength( this._filterEls )==0 ){
			var doc = this._gtd.gmail.canvasDocument;
			this._elFiltersBar.innerHTML = "";
			for( f in this._filters ){
				var el = doc.createElement("LI");
				el.className = "gtdi-filter-" + f;
				el.setAttribute("gtdi-filter-key", f);
				el.innerHTML = "<a href='#'><b><span class='gtdi-filtername'>" + glI18N.getString("ThreadListContainer.Filter." + f) + "</span><span class='gtdi-count'></span></b></a>";
				this._elFiltersBar.appendChild( el );
				this._filterEls[f] = el;
			}
		}
		
		// Refresh show/hide status:
		for( f in this._filterEls ){
			if( !this._filterEls[f].getAttribute('gtdi-hidden') && (!this._conditions[f] || this._conditions[f]()) ){
				this._filterEls[f].style.display = "";
			}else{
				this._filterEls[f].style.display = "none";
			}
		}
				
	},
	
	/**
	* Set the filter function for threads; and ensure thread visibility is set.
	* @function ?
	* @param {String} key - A unique ID for the filter function
	* @param {Function} filter - Function that accepts a thread and returns true if it should be retained
	*/
	setFilter: function(key, filter){
		this._threadList.setFilter(key, filter);
		this._refreshCounts();
	},
	
	/**
	* Get a filter for a specific key
	* @function {Function} ?
	* @param {String} key - A unique ID for the filter function
	* @return The filter function
	*/
	getFilter: function(key){
		return this._threadList.getFilter(key);
	},
	
	
	/**
	* Hide any threads that do not contain, somewhere, the search text.
	* For speed, does not consider any gtdTLThreads that have already been filtered by this._filters
	* @function ?
	* @param {String} searchText - the text to find in threads. For threads that have it, leave them showing
	*/
	search: function(searchText){
		this._threadList.search(searchText);
	},
	
	/**
	* Set the type (specified in gtdExplorerThreadsViewer), which dictates which filter buttons appear in the filter menu.
	* Set filter to 'All'
	* @function ?
	* @param {String} type - The type; specified as static properties in gtdExplorer
	* @param {String} tabName - The name of the tab currently selected
	*/
	setType: function(type, tabName){		
		// Show all
		for( f in this._filterEls ){
			if( !this._conditions[f] || this._conditions[f]() ){
				this._filterEls[f].style.display = "";
				this._filterEls[f].setAttribute('gtdi-hidden', '');
			}
		}
				
		// Hide non-appropriate filters
		var hide = null;
		switch(type){
			case gtdExplorerThreadsViewer.STATUSES:
				hide = ['next_action', 'action', 'waiting_on'];
				break;
		}
		switch(tabName){
			case gtdExplorerBundle.ALL:
			case gtdExplorerBundle.TASKS_WAITING_ON:
			case gtdExplorerBundle.TASKS_SOME_DAY:
				hide = ['next_action', 'action', 'waiting_on'];
				break;
			case gtdExplorerBundle.TASKS_TO_DO:
				hide = ['waiting_on'];
				break;
			case gtdExplorerBundle.RESOURCES:
				//hide = ['resources'];
				break;
		}
		if( hide ) for( var i = 0; i < hide.length; i++ ){
			this._filterEls[hide[i]].style.display = "none";
			this._filterEls[hide[i]].setAttribute('gtdi-hidden', '1');
		}		
		
		// Set default:
		this._setContainerFilter('all');
	},
	
	/**
	* Decide intensity with which it uses resources. If in view, set active.
	* @function {Boolean} ?
	* @param {Boolean} v - If true, set active
	* @return If v is omitted, return active status
	*/
	active: function(v){
		return this._threadList.active(v);
	},
	
	/**
	* Set how the threads are rendered. List or Project.
	* @function ?
	* @param {String} id - List or Project, using static properties of gtdExplorerViewSwitch
	*/
	setView: function(id){
		this._threadList.setView(id);
	},
	
	/**
	* Respond to a click of the filter button menu.
	* Detect filter type and apply it to the thread collection
	* @function ?
	*/
	_clickFiltersBar: function(event){
		try{
			var el = event.target;
			
			// Find out which button was clicked:
			while( el.parentNode && !el.getAttribute("gtdi-filter-key") ) el = el.parentNode;
			var key = el.getAttribute("gtdi-filter-key");
			if( key ){
				glBase.stopEvent(event);
				this._setContainerFilter( key );
			}
		}catch(e){
			glLogger.warn(e, "gtdThreadListContainer._clickFiltersBar", this._gtd.gmail);
		}
	},
	
	/** 
	* Mark previous 'active' item unactive; and set filter matching 'name' to active
	* @function ?
	* @param {String} name - The filter name to mark active
	*/
	_setContainerFilter: function(name){		
		for( f in this._filterEls ){
			if( f==name ){
				glUIWidget.addClassName( this._filterEls[f], "gtdi-active" );
				this.setFilter('filter_button', this._filters[f]);
			}else{
				glUIWidget.removeClassName( this._filterEls[f], "gtdi-active" );
			}
		}
	},
	
	/**
	* Refresh counts for each button
	* @function ?
	*/
	_refreshCounts: function(){
		// TODO can optimise by asking threadList for a filter.toString() hash of active filters; so we can abort if the filter hash hasn't changed since last time we refreshed. (Would need override for thread updates).
		for( f in this._filters ){
			this._setFilterCount( f, this._threadList.getFilterCount( this._filters[f], 'filter_button' ) );
		}
	},
	
	/**
	* Set the count figure for a filter with the given name
	* @function ?
	* @param {String} name - the filter key
	* @param {int} count - the number of threads matching the filter
	*/
	_setFilterCount: function(name, count){
		var elCount = this._filterEls[name].ownerDocument.evaluate( ".//*[@class='gtdi-count']", this._filterEls[name], null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
		elCount.innerHTML = count;
	},
	
	/**
	* Notification when the threads loaded/change, from the gtdThreadContainer in gtdThreadList. Causes refresh of counts in filter button menu.
	* @function ?
	*/
	_threadsChanged: function(event){
		this._refreshCounts();
		if( this.hasEventListener(event.name)) this.dispatchEvent(event); // propogate
	},
	
	/**
	* Notification when a thread is selected in the thread list. Propogate it.
	* @function ?
	*/
	_selectThread: function(event){
		if( this.hasEventListener(event.name)) this.dispatchEvent(event); // propogate
	},

	/**
	* Notification that a thread checkbox has been clicked
	* @function ?
	* @param {Object} event - event = {thread:, checked:true/false}
	*/
	_checkedThreads: function(event){
		// Get the checked threads - TODO, could keep a stack in memory and add/remove items
		var checkedThreads = this._threadList.getCheckedThreads();
		// Pass to Labels menu
		this._labelsMenu.loadThreads(checkedThreads);
	},
	
	/**
	* Get the threadList object 
	* @function {gtdThreadList} ?
	* @return The contained gtdThreadList object
	*/
	getThreadList: function(){
		return this._threadList;
	},
	
	/**
	* Set method for updating server. Ensure future created objects can use this info; and existing objects are updated.
	* @function ?
	* @param {String} type - The type. A static variable of glActionManager
	*/
	setServerActionType: function(type){
		this._labelsMenu.setServerActionType(type);
	},
	
	/**
	* Whether the UI object can be cached by {@link gtdMainPagePlugin.getUIObject}
	* @function {Boolean} ?
	* @return Whether the UI can be cached
	*/
	isCacheable: function() {
		return false;
	},
	

},{
	UPDATE_FILTERS: 'updateFilters',

	
	
	
});
gtdThreadListContainer.implement(glEventDispatcher); // Add event dispatching support