/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * The main class for the extension.
 * glMain is responsible for tracking page loads, and if there are plugins that
 * need to be loaded, loading and managing their instantiation and destruction.
 *
 * <h3>Developers:</h3>
 * If you'd like to learn more about plugins and the plugin API, information about
 * that can be found in {@link glPlugins glPlugins}. For more information about
 * the glGmail API start from {@link glGmail glGmail} and work out from there to
 * the various other components.
 *
 * @class ?
 */
var glMain = glBase.extend(null, {
    // Static Properties {
    /**
     * Stores all page instances currently loaded in the browser. Used to prevent
     * redundant instances and make garbage collection easier.
     * @property {private static} ?
     */
    _pageInstances: [],
    //}

    // Static Methods {
    /**
     * glMain main initializer, which is called once on Firefox load for a window.
     * Sets up page load event listeners.
     * @function {public static} ?
     */
    initialize: function() {
        glLogger.log('Initializing...', 'glMain');

        // Attach page load and unload listeners
        var appcontent = document.getElementById("appcontent");   // browser
        appcontent.addEventListener("DOMContentLoaded", glMain._pageLoad, true);
	
	// Sync the latest settings for each account:
	// TODO

        glLogger.log('Initialized', 'glMain');
    },

    /**
     * Called when a browser page is loaded. It tries to determine if it's a 
     * gmail page load and then determines if there are plugins registered for
     * that type of gmail page load. If there are, it notifies them of the page
     * gmail account and then it instantiates a glGmail object and passes it to
     * them.
     * @function {private static} ?
     * @param {Event} evt - The DOMContentLoaded event
     */
    _pageLoad: function(evt) {
        /*
         * This function is run every page load by Firefox, so it should filter
         * out unimportant page loads as quickly as possible
         */
  
  
        var doc = evt.originalTarget;
        
        try{
            // Is it roughly a gmail page?
            var url = (doc && doc.location && doc.location.href) ? doc.location.href : '';
            
            if(!/^http(s?):\/\/mail.google.com/.test(url)) return;
            
            // Are we working with a real page load?
            
            var browsers = gBrowser.browsers;
            var found = false;
            for(var i = browsers.length - 1; i>=0; i--) {
                if(doc == browsers[i].contentDocument){
                    found = true;
                    break;
                }
            }
            if(!found) return;
          
          
            // Is this page already registered?
            for(var i = glMain._pageInstances.length - 1; i>=0; i--) {
                if(glMain._pageInstances[i].doc == doc) {
                    glLogger.log('Page already registered', 'glMain');
                    return;
                }
            }
            
    
            if(glGmail.isMainGmailPageUrl(url)) {
                // Get gmail account
                var account = doc.defaultView.wrappedJSObject.top.GLOBALS[10];
                
                
                // Load account data:
                var accountData = glAccountData.retrieveForAccount(account);
                accountData.load();
                
		if( accountData.getPref('disabled') ){
		    // It is disabled for this account. No more processing.
		    accountData.shutdown();
		    return;
		}
		
		var plugins = new glPlugins(accountData);
		if( plugins.getLoadedPlugins().length==0 ){
		    // No plugins loaded
		    glLogger.log('account has no plugins', 'glMain._loadAccount');
		    accountData.shutdown();
		    plugins.unload();
		    return;
		}
				
		// Create the glGmail instance
		var gmailInstance = null;
		try {
		    gmailInstance = new glGmail(doc, accountData, plugins);
		} catch(e) {
		    glLogger.warn('Problem creating glGmail instance'+"\n"+e, 'glMain', null, doc);
		    accountData.shutdown();
		    plugins.unload();
		    return; // Can't continue if glGmail can't be instantiated
		}
	    
		
		// Load the plugins for this account:
		plugins.loadPluginsForGmail(gmailInstance);
		
		// Made it to end, so save as successful page instance and register page close handler - YEAH!
		glMain._pageInstances.push({
		    doc: doc,
		    gmail: gmailInstance,
		    plugins: plugins,
		    accountData: accountData,
		});
		doc.defaultView.addEventListener('pagehide', glMain._pageUnload, true);
                
            } else {
                glLogger.log('Page not currently supported or not a gmail page: '+url, 'glMain');
            }
        } catch(e) {
            glLogger.warn('Problem loading page'+"\n"+e, 'glMain', null, doc);
        }
    
    },


    /**
     * Called when a browser page is unloaded. Sees if there is a pageInstance
     * object for the given document, and if there is, unloads all plugins and
     * the glGmail object.
     * @function {private static} ?
     * @param {Event} evt - The pagehide event
     */
    _pageUnload: function(evt) {
        var doc = evt.target.ownerDocument != null ? evt.target.ownerDocument : evt.target;
        if(!!doc.defaultView.frameElement || doc.URL == "about:blank") return;
        doc.defaultView.removeEventListener('pagehide', glMain._pageUnload, true);

        for(var i = 0; i < glMain._pageInstances.length; i++) {
            var page = glMain._pageInstances[i];
            if(page.doc == doc) {
                // It's an active page so shut everything down and clean up
                glLogger.log('Unloading active page...', 'glMain');
                try {
                    page.accountData.shutdown();
                    page.plugins.unload();
                    page.gmail.unload();
                } catch(e) {
                    glLogger.warn('Error during shutdown:'+'\n'+e, 'glMain._pageUnload'); 
                }

                glMain._pageInstances.splice(i, 1);
                return;
            }
        }
    }
    //}
});


// Run glMain.initialize() when ready and remove listener
var _glLoadFunc = function() {
    window.removeEventListener('load', _glLoadFunc, true);
    glMain.initialize();
};
window.addEventListener('load', _glLoadFunc, true);
