/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */



/**
 * Event Dispatcher Mixin.
 * Gives the extending class event dispatching capabilities.
 *
 * @interface ?
 */
var glEventDispatcher = glBase.extend({
    // Properties {
    /**
     * Event listener store
     * @property {Array} ?
     */
    __listeners: null,
    //}

    // Methods {
    /**
     * Registers an event listener with the EventDispatcher so that it may receive event notifications
     * @function ?
     * @param {String} eventName - The event to listen for
     * @param {Function} listener - The function to call on event dispatch
     * @param {optional Object} thisObj - The optional object to use as "this" when the event is dispatched
     */
    addEventListener: function(eventName, listener, thisObj) {
        if(this.__listeners == null) this.__listeners = {};
        if(!this.__listeners[eventName]) this.__listeners[eventName] = [];

        if(!listener) glLogger.error('No listener', 'glEventDispatcher');

        // Check and see if it hasn't been added before adding
        var found = false;
        for each(var l in this.__listeners[eventName]) {
            if(l.func === listener && l.thisObj==thisObj ) {
                found = true;
                break;
            }
        }
        if(!found) this.__listeners[eventName].push({func: listener, thisObj: thisObj});
		return !found;
    },

    /**
     * Removes the given listener from the EventDispatcher
     * @function ?
     * @param {String} eventName - The event being listened for
     * @param {Function} listener - The function that was registered as a listener
     */
    removeEventListener: function(eventName, listener, thisObj) {
        if(!this.hasEventListener(eventName)) return;

        for(var i = 0; i < this.__listeners[eventName].length; i++) {
            if(this.__listeners[eventName][i].func === listener && this.__listeners[eventName][i].thisObj==thisObj) {
                this.__listeners[eventName].splice(i, 1);
                break;
            }
        }
    },

    /**
     * Dispatches the given event to all listeners
     * @function ?
     * @param {Event} event - The event to dispatch
     */
    dispatchEvent: function(event) {
        if(this.hasEventListener(event.name)) {
            // Add target property to event if it doesn't already exist
            if(typeof event['target'] == 'undefined') {
                event['target'] = this;
            }

            for each(var l in this.__listeners[event.name]) {
            	try{	// Catch errors and neutralise them; so one bad event does not bring down all listening.
                	l.func.apply(l.thisObj, [event]);
                }catch(e){
                	glLogger.warn(e, 'glEventDispatcher.dispatchEvent');
                }
            }
        }
    },

    /**
     * Checks whether the EventDispatcher has any listeners registered for the
     * given event.
     * @function {Boolean} ?
     * @param {String} eventName - The event to check for listeners for
     * @return Whether there are any event listeners registered
     */
    hasEventListener: function(eventName) {
        if(this.__listeners && this.__listeners[eventName] && this.__listeners[eventName].length) {
            return true;
        } else {
            return false;
        }
    },

    /**
     * Removes all listeners from the object. Can be used for clean up when you
     * don't want them to be registered any more.
     * @function {protected} ?
     */
    _removeAllListeners: function() {
        this.__listeners = null;
    },
    //}
});
